#include "buf.h"

#include <stdlib.h>
#include <string.h>
#include <assert.h>

/****************************************************************************
 * LINE BUFFER
 */

struct buf_line {
	char *buf;
	int alloc;
	int at;

	void *data;
	buf_line_call_t cb;
};

/* create new line buffer */
buf_line_t *buf_line_new(void *data, buf_line_call_t cb)
{
	buf_line_t *lb = malloc(sizeof *lb);
	if(!lb) abort();

	lb->buf = NULL;
	lb->alloc = 0;
	lb->at = 0;

	lb->data = data;
	lb->cb = cb;

	return lb;
}

/* change data and callback */
void buf_line_setcb(buf_line_t *lb, void *data, buf_line_call_t cb)
{
	lb->data = data;
	lb->cb = cb;
}

/* free line buffer */
void buf_line_free(buf_line_t *lb)
{
	free(lb->buf);
	free(lb);
}

/* write data into line buffer
 * callback gets called with new lines
 * returns number of new lines
 * EOLs are cut off
 * final \0 (in place of EOL) is guaranteed */
int buf_line_write(buf_line_t *lb, const char *buf, int len)
{
	if(!len) return 0;

	if(lb->at + len + 8 > lb->alloc) {
		lb->alloc = lb->at + len + 8;
		lb->buf = realloc(lb->buf, lb->alloc);
		if(!lb->buf) abort();
	}

	memcpy(lb->buf + lb->at, buf, len);

	int lastchk = lb->at;
	int last = 0;
	int lines = 0;
	int wineol;

	lb->at += len;

	while(lastchk < lb->at) {
		char *eol = memchr(lb->buf + lastchk, '\n', lb->at - lastchk);
		if(!eol) break;

		wineol = 0;
		if(eol > lb->buf + lastchk) {
			if(*(eol-1) == '\r') {
				*eol = 0;
				eol --;
				wineol = 1;
			}
		}

		*eol = 0;
		lb->cb(lb->data, lb->buf + last, eol - (lb->buf + last));
		lines ++;

		last = (eol - lb->buf) + 1 + wineol;
		lastchk = last;
	}

	if(lines > 0) {
		if(lastchk < lb->at) {
			memmove(lb->buf, lb->buf + lastchk, lb->at - lastchk);
		}

		lb->at -= lastchk;
	}

	return lines;
}

/****************************************************************************
 * FIFO
 */

typedef struct buf_packet buf_packet_t;

struct buf_packet {
	buf_packet_t *next;
	char data[];
};

struct buf_fifo {
	int size;  /* packet size */

	buf_packet_t *first;
	buf_packet_t *last;

	int at; /* offset in first packet */
	int len; /* length of data in last packet */
};

/* create new fifo */
buf_fifo_t *buf_fifo_new(int packet)
{
	buf_fifo_t *bf = malloc(sizeof *bf);
	if(!bf) abort();

	bf->size = packet;

	buf_packet_t *bp = malloc(packet + sizeof *bp);
	if(!bp) abort();
	bp->next = NULL;

	bf->first = bp;
	bf->last = bp;
	bf->at = 0;
	bf->len = 0;

	return bf;
}

/* free fifo */
void buf_fifo_free(buf_fifo_t *bf)
{
	buf_packet_t *bp = bf->first;
	while(bp) {
		buf_packet_t *f = bp;
		bp = bp->next;
		free(f);
	}

	free(bf);
}

/* write into fifo */
void buf_fifo_write(buf_fifo_t *bf, const char *buf, int len)
{
	while(len > 0) {
		int bite = len;
		if(bite > bf->size - bf->len) bite = bf->size - bf->len;

		memcpy(bf->last->data + bf->len, buf, bite);
		bf->len += bite;
		buf += bite;
		len -= bite;

		if(bf->len == bf->size) {
			bf->len = 0;

			buf_packet_t *bp = malloc(bf->size + sizeof *bp);
			if(!bp) abort();
			bp->next = NULL;
			bf->last->next = bp;
			bf->last = bp;
		}
	}
}

/* get next available packet size
 * if 0, fifo is empty */
int buf_fifo_packetsize(buf_fifo_t *bf)
{
	if(bf->last != bf->first) return bf->size - bf->at;

	return bf->len - bf->at;
}

/* get next available packet
 * writes pointer to packet to buf
 * returns packet size
 * packet is not removed from fifo */
int buf_fifo_getpacket(buf_fifo_t *bf, char **buf)
{
	*buf = bf->first->data + bf->at;

	if(bf->last != bf->first) return bf->size - bf->at;
	return bf->len - bf->at;
}

/* remove bytes from fifo
 * (cannot be more than the available packet) */
void buf_fifo_advance(buf_fifo_t *bf, int n)
{
	if(bf->last == bf->first) {
		assert(n <= bf->len - bf->at);

		bf->at += n;
		if(bf->at == bf->len) {
			bf->at = 0;
			bf->len = 0;
		}
	} else {
		assert(n <= bf->size - bf->at);

		bf->at += n;
		if(bf->at == bf->size) {
			bf->at = 0;
			buf_packet_t *bp = bf->first;
			bf->first = bf->first->next;
			free(bp);
		}
	}
}
