#include "parse.h"

#include <string.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <math.h>

const char *parse_error(parse_result_t res)
{
	switch(res) {
	case PARSE_OK:
		return "OK";
	case PARSE_UNREC:
		return "Command not recognized";
	case PARSE_DEF:
		return "Keyword after 'default' not recognized";
	case PARSE_TERM:
		return "Invalid termination action";
	case PARSE_ID:
		return "Invalid identifier (needs to be [a-zA-Z_][a-zA-Z_0-9]*)";
	case PARSE_SHORT:
		return "Short statement (ran out of tokens)";
	case PARSE_FD:
		return "Invalid FD specification";
	case PARSE_REC:
		return "Invalid record size";
	case PARSE_PROCEQ:
		return "Expected = in proc statement, got something else";
	case PARSE_LONG:
		return "Unexpected tokens at end of statement";
	case PARSE_BIN:
		return "Invalid syntax for hex binary data";
	case PARSE_CALL:
		return "Parse callback returned nonzero";
	case PARSE_TIME:
		return "Invalid timer delay spec";
	}

	abort();
}

#define TERM_INVALID 65536

static termination_t parse_term(char *s)
{
	if(!strcmp(s, "kill"))
		return TERM_KILL;
	else if(!strcmp(s, "event"))
		return TERM_EVENT;

	return TERM_INVALID;
}

static int parse_cascade(char *s)
{
	if(!strcmp(s, "keep"))
		return 0;
	else if(!strcmp(s, "cascade"))
		return 1;

	return -1;
}

/* check whether string is a valid identifier */
static int valid_id(const char *id)
{
	if(!*id) return 0;
	if(!isalpha(*id) && *id != '_') return 0;
	id ++;
	for(; *id; id ++)
		if(!isalnum(*id) && *id != '_') return 0;
	return 1;
}

/* parse FD */
static int parse_fd(const char *s)
{
	char *end;
	int fd;

	if(!*s) return -1;
	fd = strtol(s, &end, 0);
	if(*end) return -1;
	return fd;
}

static int unhex(char c)
{
	if(c >= '0' && c <= '9') return c - '0';
	c = toupper(c);
	if(c >= 'A' && c <= 'F') return c - 'A' + 10;
	return -1;
}

/* unpack binary data given in hex
 * writes back in same buffer, returns length */
static int unpackbin(char *buf)
{
	char *p = buf;
	char *q = buf;
	int minor, major;

	while(*buf) {
		major = unhex(*buf++);
		if(major<0) return -1;
		minor = unhex(*buf++);
		if(minor<0) return -1;
		*q++ = major*16 + minor;
	}

	return q - p;
}

static p_proc_infd_t *infds = NULL;
static int infdn = 0;
static int infd_alloc = 0;
static p_proc_outfd_t *outfds = NULL;
static int outfdn = 0;
static int outfd_alloc = 0;

static void resetprocfds()
{
	infdn=0;
	outfdn=0;
}

static p_proc_infd_t *addinfd()
{
	if(infdn >= infd_alloc) {
		infd_alloc += 8;
		infds = realloc(infds, infd_alloc * sizeof *infds);
		if(!infds) abort();
	}

	return infds + (infdn ++);
}

static p_proc_outfd_t *addoutfd()
{
	if(outfdn >= outfd_alloc) {
		outfd_alloc += 8;
		outfds = realloc(outfds, outfd_alloc * sizeof *outfds);
		if(!outfds) abort();
	}

	return outfds + (outfdn ++);
}

static int nextpipeid = 0;
static int nexttimerid = 0;

parse_result_t parse(void *data, int argc, char **argv)
{
	int a = 0, pause, r, fd, endchain, bin, pfx, repeat;
	char *cmd, *id;
	char *recsize, *end, *name;
	char pipeid[20];
	char timerid[20];
	char *srcdef, *tgtdef;
	char *sep, *type;

	struct timeval delay;
	double delay_s;

	termination_t term;

	p_proc_infd_t *infd;
	p_proc_outfd_t *outfd;
	p_pipe_src_t src;
	p_pipe_tgt_t tgt;
	p_pipe_place_t place;

	filter_type_t filter;
	char *filterdata;
	int filterlen;

	char *prefixdata;
	int prefixlen;

	if(!argc) return PARSE_OK;

#define PEEK ({ if(a >= argc) return PARSE_SHORT; argv[a]; })
#define NEXT ({ if(a >= argc) return PARSE_SHORT; argv[a++]; })
#define NEXTID ({ char *id = NEXT; if(!valid_id(id)) return PARSE_ID; id; })
#define NEXTTERM \
	({ termination_t t = parse_term(NEXT); \
	 if(t == TERM_INVALID) return PARSE_TERM; \
	 t; })

#define NEXTFD \
	({ char *fds = NEXT; \
	 int fd = parse_fd(fds); \
	 if(fd < 0) return PARSE_FD; \
	 fd; })

#define ENOUGH if(a < argc) return PARSE_LONG;

	cmd = NEXT;

	if(!strcmp(cmd, "endproc")) {
		if(p_endproc(data, NEXTID)) return PARSE_CALL;
		ENOUGH;
		return PARSE_OK;
	}

	if(!strcmp(cmd, "endpipe")) {
		if(p_endpipe(data, NEXTID)) return PARSE_CALL;
		ENOUGH;
		return PARSE_OK;
	}

	if(!strcmp(cmd, "unpause")) {
		if(p_unpause(data, NEXTID)) return PARSE_CALL;
		ENOUGH;
		return PARSE_OK;
	}

	if(!strcmp(cmd, "set")) {
		if(p_set(data, NEXTID, argc-2, argv+2)) return PARSE_CALL;
		return PARSE_OK;
	}

	if(!strcmp(cmd, "default")) {
		name = NEXT;
		if(!strcmp(name, "proc")) {
			if(p_default_proc_term(data, NEXTTERM)) return PARSE_CALL;
			ENOUGH;
			return PARSE_OK;
		}

		if(!strcmp(name, "eof")) {
			if(p_default_eof_term(data, NEXTTERM)) return PARSE_CALL;
			ENOUGH;
			return PARSE_OK;
		}

		if(!strcmp(name, "write")) {
			if(p_default_write_term(data, NEXTTERM)) return PARSE_CALL;
			ENOUGH;
			return PARSE_OK;
		}

		return PARSE_DEF;
	}

	if(!strcmp(cmd, "proc")) {
		id = NEXTID;
		term = parse_term(PEEK);
		if(term == TERM_INVALID) {
			term = TERM_DEFAULT;
		} else {
			a ++;
		}

		resetprocfds();

		pause = 0;
		if(!strcmp(PEEK, "-pause")) {
			pause = 1;
			a ++;
		}

		while(!strcmp(PEEK, "-i")) {
			a ++;

			infd = addinfd();

			infd->fd = NEXTFD;
			infd->term = parse_term(PEEK);
			if(infd->term == TERM_INVALID) {
				infd->term = TERM_DEFAULT;
			} else {
				a ++;
			}

			infd->cascade = parse_cascade(PEEK);
			if(infd->cascade != -1) a ++;
		}

		while(!strcmp(PEEK, "-o")) {
			a ++;

			outfd = addoutfd();

			outfd->fd = NEXTFD;
			outfd->record = -1;

			recsize = PEEK;
			r = strtol(recsize, &end, 0);
			if(*recsize && !*end) {
				if(r < 0) return PARSE_REC;
				outfd->record = r;
				a ++;
			}

			outfd->term = parse_term(PEEK);
			if(outfd->term == TERM_INVALID) {
				outfd->term = TERM_DEFAULT;
			} else {
				a ++;
			}

			outfd->cascade = parse_cascade(PEEK);
			if(outfd->cascade != -1) a ++;
		}

		if(strcmp(NEXT, "=")) return PARSE_PROCEQ;

		PEEK;

		if(p_proc(data, id, term, pause, infdn, infds, outfdn, outfds, argc-a, argv+a)) return PARSE_CALL;

		return PARSE_OK;
	}

	if(!strcmp(cmd, "timer")) {
		nexttimerid ++;
		sprintf(timerid, "__timer_%d", nexttimerid);
		id = timerid;

		if(!strcmp(PEEK, "as")) {
			a ++;
			id = NEXTID;
		}

		repeat = 0;
		if(!strcmp(PEEK, "repeat")) {
			a ++;
			repeat = 1;
		}

		errno = 0;
		delay_s = strtod(NEXT, &end);
		if(errno || *end) return PARSE_TIME;

		if(delay_s < 0.0) return PARSE_TIME;

		delay.tv_sec = (int) delay_s;
		delay.tv_usec = (delay_s - floor(delay_s)) * 1000000.0;

		end = NEXT;

		ENOUGH;

		if(p_timer(data, id, repeat, delay, end)) return PARSE_CALL;

		return PARSE_OK;
	}

	if(!strcmp(cmd, "endtimer")) {
		if(p_endtimer(data, NEXTID)) return PARSE_CALL;
		ENOUGH;
		return PARSE_OK;
	}

	if(!strcmp(cmd, "pipe")) {
		nextpipeid ++;
		sprintf(pipeid, "__pipe_%d", nextpipeid);
		id = pipeid;

		if(!strcmp(PEEK, "as")) {
			a ++;
			id = NEXTID;
		}

		place = PPLACE_LAST;

		if(!strcmp(PEEK, "-front")) {
			a ++;
			place = PPLACE_FIRST;
		}

		src.proc = NULL;
		src.fd = -1;

		srcdef = NEXT;
		if(!strcmp(srcdef, ":ev")) {
			src.type = PSRC_EV;
		} else {
			char *sep = strchr(srcdef, '/');
			int fd = -1;

			if(sep) {
				*sep = 0;
				if(!sep[1]) return PARSE_FD;
				fd = strtol(sep+1, &end, 0);
				if(*end || fd < 0) return PARSE_FD;
			}

			if(!strcmp(srcdef, ":plumb")) {
				src.type = PSRC_PLUMB;
				src.fd = (fd == -1) ? 0 : fd;
			} else {
				src.type = PSRC_PROC;
				if(!valid_id(srcdef)) return PARSE_ID;
				src.proc = srcdef;
				src.fd = (fd == -1) ? 1 : fd;
			}
		}

		tgt.proc = NULL;
		tgt.fd = -1;

		tgtdef = NEXT;
		if(!strcmp(tgtdef, ":ctrl")) {
			tgt.type = PTGT_CTRL;
		} else {
			sep = strchr(tgtdef, '/');
			fd = -1;

			if(sep) {
				*sep = 0;
				if(!sep[1]) return PARSE_FD;
				fd = strtol(sep+1, &end, 0);
				if(*end || fd < 0) return PARSE_FD;
			}

			if(!strcmp(tgtdef, ":plumb")) {
				tgt.type = PTGT_PLUMB;
				tgt.fd = (fd == -1) ? 1 : fd;
			} else {
				tgt.type = PTGT_PROC;
				if(!valid_id(tgtdef)) return PARSE_ID;
				tgt.proc = tgtdef;
				tgt.fd = (fd == -1) ? 0 : fd;
			}
		}

		filter = FILT_NONE;
		filterdata = NULL;
		filterlen = 0;

		prefixdata = NULL;
		prefixlen = 0;

		if(a < argc) {
			type = PEEK;
			bin = 0;
			if(!strcmp(type, "-filter"))
				filter = FILT_FILTER;
			else if(!strcmp(type, "-cut"))
				filter = FILT_CUT;
			else if(!strcmp(type, "-filter:hex")) {
				filter = FILT_FILTER;
				bin = 1;
			} else if(!strcmp(type, "-cut:hex")) {
				filter = FILT_CUT;
				bin = 1;
			}

			if(filter != FILT_NONE) {
				a ++;
				filterdata = NEXT;
				if(bin) {
					filterlen = unpackbin(filterdata);
					if(filterlen < 0) return PARSE_BIN;
				} else {
					filterlen = strlen(filterdata);
				}
			}
		}

		if(a < argc) {
			type = PEEK;
			bin = 0;
			pfx = 0;

			if(!strcmp(type, "-prefix:hex")) {
				pfx = 1;
				bin = 1;
			} else if(!strcmp(type, "-prefix")) {
				pfx = 1;
			}

			if(pfx) {
				a ++;
				prefixdata = NEXT;
				if(bin) {
					prefixlen = unpackbin(prefixdata);
					if(prefixlen < 0) return PARSE_BIN;
				} else {
					prefixlen = strlen(prefixdata);
				}
			}
		}

		endchain = 0;
		if(a < argc) {
			if(!strcmp(PEEK, "-endchain")) {
				endchain = 1;
				a ++;
			}
		}

		if(p_pipe(data, id, src, tgt, filter, filterlen, filterdata, prefixlen, prefixdata, endchain, place)) return PARSE_CALL;

		ENOUGH;

		return PARSE_OK;
	}

	return PARSE_UNREC;
}
