#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <assert.h>

#include <errno.h>

#include <sys/types.h>
#include <signal.h>

#include <libgen.h> /* basename */

#include "plumb.h"
#include "tokens.h"

/* our main tokenizer */
tok_t *tkn = NULL;

/* default proc termination action */
termination_t term_proc = TERM_KILL;

/* default eof termination action */
termination_t term_eof = TERM_KILL;

/* default write error termination action */
termination_t term_write = TERM_KILL;

/* default src cascade */
int default_src_cascade = 1;
/* default tgt cascade */
int default_tgt_cascade = 1;

char *proc_kill_reason_str[] = {
	"KILL",
	"CASCADE",
	"EXIT"
};

char *pipe_kill_reason_str[] = {
	"KILL",
	"EOF",
	"READ",
	"WRITE",
	"CASCADE"
};

char *src_end_reason_str[] = {
	"KILL",
	"EOF",
	"READ",
	"CASCADE"
};

char *tgt_end_reason_str[] = {
	"KILL",
	"WRITE",
	"CASCADE"
};

#ifdef WIN32
#include "plumb_os_win32.h"
#else
#include "plumb_os_unix.h"
#endif

/****************************************************************************
 * log
 ****************************************************************************/

/* current main log level */
log_level_t loglevel = LOG_ERROR;

void plog(log_level_t lev, const char *fmt, ...)
{
	int len;
	va_list ap;
	char *level;

	if(lev < loglevel) return;

	switch(lev) {
	case LOG_ERROR:
		level = "ERROR";
		break;
	case LOG_WARNING:
		level = "WARNING";
		break;
	case LOG_DEBUG:
		level = "DEBUG";
	case LOG_JUNK:
		level = "JUNK";
		break;
	}

	/* write log to stderr */

	va_start(ap, fmt);
	len = vbufprintf(fmt, ap);
	va_end(ap);
	fwrite(bufprintfbuf, len, 1, stderr);
	fflush(stderr);

/* don't log to global_evsrc for the moment.....

	if(lev < global_evsrc->level) return;

	len += strlen(level) + 1;

	static char *evbuf = NULL;
	static int evalloc = 0;

	if(len+1 > evalloc) {
		evalloc = len+1;
		evbuf = realloc(evbuf, evalloc);
		if(!evbuf) abort();
	}

	sprintf(evbuf, "%s ", level);
	strcat(evbuf, bufprintfbuf);

	src_write((src_t *) global_evsrc, evbuf, len);
*/
}

/****************************************************************************
 * parser callbacks
 ****************************************************************************/

/* callback for tokenizer that gives result to plumb parser */
int parse_line(int tc, char *tv[])
{
	parse_result_t res = parse(NULL, tc, tv);

	if(res == PARSE_CALL) return PARSE_CALL;

	if(res) {
		plog(LOG_ERROR, "Parse error (%s) in statement: %s\n", parse_error(res), tok_print(tc, tv));
		return res;
	}

	return 0;
}

int p_set(void *data, char *key, int argc, char **argv)
{
	plog(LOG_DEBUG, "SET variable %s = %s\n", key, tok_print(argc, argv));
	tok_setvar(tkn, key, argc, argv);
	return 0;
}

int p_default_proc_term(void *data, termination_t term)
{
	plog(LOG_DEBUG, "DEFAULT PROC TERM %d\n", term);
	term_proc = term;
	return 0;
}

int p_default_eof_term(void *data, termination_t term)
{
	plog(LOG_DEBUG, "DEFAULT EOF TERM %d\n", term);
	term_eof = term;
	return 0;
}

int p_default_write_term(void *data, termination_t term)
{
	plog(LOG_DEBUG, "DEFAULT WRITE TERM %d\n", term);
	term_write = term;
	return 0;
}

int p_default_src_cascade(void *data, int cascade)
{
	plog(LOG_DEBUG, "DEFAULT SOURCE CASCADE %d\n", cascade);
	default_src_cascade = cascade;
	return 0;
}

int p_default_tgt_cascade(void *data, int cascade)
{
	plog(LOG_DEBUG, "DEFAULT TARGET CASCADE %d\n", cascade);
	default_tgt_cascade = cascade;
	return 0;
}

int p_proc(void *data, const char *id, termination_t term, int pause, int infdn, const p_proc_infd_t *infds, int outfdn, const p_proc_outfd_t *outfds, int argc, char **argv)
{
	int i;

	plog(LOG_DEBUG, "PROC %s = %s\n", id, tok_print(argc, argv));

	ptrhash_entry_t *pte = ptrhash_getentry(procs, (char *) id);
	if(pte) {
		/* process with this ID exists already */
		plog(LOG_ERROR, "Cannot create process '%s' (with command line %s), because it exists already!\n", id, tok_print(argc, argv));
		return 1;
	}

	/* create new process record */

	proc_t *p = malloc(sizeof *p);
	if(!p) abort();

	p->killing = 0;

	p->id = strdup(id);
	ptrhash_set(procs, p->id, p);
	nprocs ++;

	ptrset_init(&p->srcs, NULL);
	ptrset_init(&p->tgts, NULL);

	p->onend = term == TERM_DEFAULT ? term_proc : term;

	p->pid = -1;

	/* copy command line */

	p->argc = argc;
	p->argv = malloc((argc+1) * sizeof(p->argv[0]));
	if(!p->argv) abort();

	for(i=0; i<argc; i++) {
		p->argv[i] = strdup(argv[i]);
	}
	p->argv[i] = NULL;

	/* create sources */

	int hasstdout = 0;
	int hasstderr = 0;

	for(i=0; i<outfdn; i++) {
		src_fd_t *s = src_fd_new(outfds[i].fd, outfds[i].record, outfds[i].term);
		ptrset_add(&p->srcs, s);
		s->src.cascade = outfds[i].cascade < 0 ? default_src_cascade : outfds[i].cascade;

		if(outfds[i].fd == 1) hasstdout = 1;
		else if(outfds[i].fd == 2) hasstderr = 1;
	}

	/* if there's no defined stdout, create one */
	if(!hasstdout) {
		src_fd_t *s = src_fd_new(1, -1, term_eof);
		ptrset_add(&p->srcs, s);
		s->src.cascade = default_src_cascade;
	}

	/* if there's no defined stderr, create one */
	if(!hasstderr) {
		src_fd_t *s = src_fd_new(2, -1, term_eof);
		ptrset_add(&p->srcs, s);
		s->src.cascade = default_src_cascade;

		/* make pipe to plumb stderr */

		plog(LOG_DEBUG, "Implicit pipe from %s/2 to :plumb/2\n", id);

		pipe_t *pi = malloc(sizeof *pi);
		if(!pi) abort();

		char pipid[32];
		static int implicit_stderr = 0;
		implicit_stderr ++;
		sprintf(pipid, "__impl_stderr_%d", implicit_stderr);

		ptrhash_entry_t *pte = ptrhash_getentry(pipes, pipid);
		if(pte) {
			plog(LOG_ERROR, "Cannot create implicit stderr pipe from %s/2, because pipe id %s exists already! (internal error)\n", id, pipid);
			endplumb();
			exit(1);
		}

		pi->id = strdup(pipid);

		ptrhash_set(pipes, pi->id, pi);
		npipes ++;

		pi->src = (src_t *) s;
		pi->tgt = global_tgts.p[1];
		pi->src_el = ptrlist_add_last(&pi->src->pipes, pi);
		ptrset_add(&pi->tgt->pipes, pi);

		pi->filter = FILT_NONE;
		pi->filterdata = NULL;
		pi->filterlen = 0;

		pi->endchain = 0;

		pi->prefixlen = 0;
		pi->prefixdata = NULL;
	}

	/* create targets */

	int hasstdin = 0;

	for(i=0; i<infdn; i++) {
		tgt_fd_t *t = tgt_fd_new(infds[i].fd, infds[i].term);
		ptrset_add(&p->tgts, t);
		t->tgt.cascade = infds[i].cascade < 0 ? default_tgt_cascade : infds[i].cascade;

		if(infds[i].fd == 0) hasstdin = 1;
	}

	/* if there's no defined stdin, create one */
	if(!hasstdin) {
		tgt_fd_t *t = tgt_fd_new(0, term_write);
		ptrset_add(&p->tgts, t);
		t->tgt.cascade = default_tgt_cascade;
	}

	if(os_fork(p, pause)) {
		plog(LOG_DEBUG, "os_fork %s returned error, calling proc_kill\n", p->id);
		proc_kill(p, PKR_KILL, 1);
		return 1;
	}

	return 0;
}

int p_pipe(void *data, const char *id, p_pipe_src_t src, p_pipe_tgt_t tgt, filter_type_t filter, int filterlen, const char *filterdata, int prefixlen, const char *prefixdata, int endchain, p_pipe_place_t place)
{
	int i;
	char *srcname, *tgtname;
	int srcfd, tgtfd;

	if(src.type == PSRC_PROC) {
		srcname = src.proc;
		srcfd = src.fd;
	} else if(src.type == PSRC_EV) {
		srcname = ":ev";
		srcfd = 0;
	} else if(src.type == PSRC_PLUMB) {
		srcname = ":plumb";
		srcfd = src.fd;
	}

	if(tgt.type == PTGT_PROC) {
		tgtname = tgt.proc;
		tgtfd = tgt.fd;
	} else if(tgt.type == PTGT_CTRL) {
		tgtname = ":ctrl";
		tgtfd = 0;
	} else if(tgt.type == PTGT_PLUMB) {
		tgtname = ":plumb";
		tgtfd = tgt.fd;
	}

	plog(LOG_DEBUG, "PIPE %s = %s/%d -> %s/%d\n", id, srcname, srcfd, tgtname, tgtfd);

	ptrhash_entry_t *pte = ptrhash_getentry(pipes, (char *) id);
	if(pte) {
		plog(LOG_ERROR, "Cannot create pipe '%s', it exists already!\n", id);
		return 1;
	}

	src_t *psrc = NULL;
	tgt_t *ptgt = NULL;

	/* look up source */

	switch(src.type) {
	case PSRC_PROC:
	case PSRC_PLUMB:
		{
			ptrset_t *srcl; /* source list to use */
			char *srcname;  /* name/ID of source */

			if(src.type == PSRC_PROC) {
				/* if it's a process, look up process by ID */
				pte = ptrhash_getentry(procs, src.proc);
				if(pte) {
					proc_t *pr = pte->value;
					srcl = &pr->srcs;
					srcname = pr->id;
				} else {
					/* proc not found is ignored because there's a race condition:
					 * starting a process then attaching pipes
					 * if the process didn't start or exited right away,
					 * attaching pipes might fail before the exit event gets to
					 * the controlling process */
					plog(LOG_ERROR, "IGNORING attempt to attach pipe %s to source %s/%d (process doesn't exist)\n", id, src.proc, src.fd);
					return 0;
				}
			} else {
				/* if it's :plumb, use global source list */
				srcl = &global_srcs;
				srcname = ":plumb";
			}

			/* look up source by FD */
			for(i=0; i<srcl->n; i++) {
				src_t *s = srcl->p[i];
				assert(s->type == SRC_FD);
				src_fd_t *sf = (src_fd_t *) s;
				if(sf->fd == src.fd) {
					psrc = s;
					break;
				}
			}
			if(i == srcl->n) {
				plog(LOG_ERROR, "Cannot attach pipe %s to source %s/%d, process doesn't have that FD as source\n", id, srcname, src.fd);
				return 1;
			}
		}
		break;
	case PSRC_EV:
		psrc = (src_t *) global_evsrc;
		break;
	}

	/* look up target */

	switch(tgt.type) {
	case PTGT_PROC:
	case PTGT_PLUMB:
		{
			ptrset_t *tgtl; /* target list to use */
			char *tgtname;  /* name/ID of target */

			if(tgt.type == PTGT_PROC) {
				/* if it's a process, look up process by ID */
				pte = ptrhash_getentry(procs, tgt.proc);
				if(pte) {
					proc_t *pr = pte->value;
					tgtl = &pr->tgts;
					tgtname = pr->id;
				} else {
					/* proc not found is ignored because there's a race condition:
					 * starting a process then attaching pipes
					 * if the process didn't start or exited right away,
					 * attaching pipes might fail before the exit event gets to
					 * the controlling process */
					plog(LOG_ERROR, "IGNORING attempt to attach pipe %s to target %s/%d (process doesn't exist)\n", id, tgt.proc, tgt.fd);
					return 0;
				}
			} else {
				/* if it's :plumb, use global target list */
				tgtl = &global_tgts;
				tgtname = ":plumb";
			}

			/* look up target by FD */
			for(i=0; i<tgtl->n; i++) {
				tgt_t *t = tgtl->p[i];
				assert(t->type == TGT_FD);
				tgt_fd_t *tf = (tgt_fd_t *) t;
				if(tf->fd == tgt.fd) {
					ptgt = t;
					break;
				}
			}
			if(i == tgtl->n) {
				plog(LOG_ERROR, "Cannot attach pipe %s to target %s/%d, process doesn't have that FD as target\n", id, tgtname, tgt.fd);
				return 1;
			}
		}
		break;
	case PTGT_CTRL:
		ptgt = (tgt_t *) global_ctrltgt;
		break;
	}

	/* create pipe record */

	if(ptgt->type == TGT_CTRL && (prefixlen || psrc->type != SRC_FD || ((src_fd_t *) psrc)->recordsize != -1)) {
		plog(LOG_ERROR, "Cannot attach pipe %s to :ctrl - source must be FD, there can be no prefix and buffering has to be line-based\n", id);
		return 1;
	}

	pipe_t *p = malloc(sizeof *p);
	if(!p) abort();

	p->id = strdup(id);
	ptrhash_set(pipes, p->id, p);

	if(psrc != (src_t *) global_evsrc) npipes ++;

	p->src = psrc;
	p->tgt = ptgt;

	switch(place) {
	case PPLACE_FIRST:
		p->src_el = ptrlist_add_first(&psrc->pipes, p);
		break;
	case PPLACE_LAST:
		p->src_el = ptrlist_add_last(&psrc->pipes, p);
		break;
	}

	ptrset_add(&ptgt->pipes, p);

	p->filter = filter;
	p->filterlen = filterlen;
	if(filterlen > 0) {
		p->filterdata = malloc(filterlen);
		if(!p->filterdata) abort();
		memcpy(p->filterdata, filterdata, filterlen);
	} else {
		p->filterdata = NULL;
	}

	p->endchain = endchain;

	p->prefixlen = prefixlen;
	if(prefixlen > 0) {
		p->prefixdata = malloc(prefixlen);
		if(!p->prefixdata) abort();
		memcpy(p->prefixdata, prefixdata, prefixlen);
	} else {
		p->prefixdata = NULL;
	}

	return 0;
}

int p_endproc(void *data, char *id)
{
	plog(LOG_DEBUG, "ENDPROC %s\n", id);

	ptrhash_entry_t *pte = ptrhash_getentry(procs, (char *) id);
	if(pte) {
		proc_kill(pte->value, PKR_KILL, 0);
		return 0;
	} else {
		plog(LOG_ERROR, "IGNORING attempt to kill process '%s' (it doesn't exist)\n", id);
		return 0;
	}
}

int p_endpipe(void *data, char *id)
{
	plog(LOG_DEBUG, "ENDPIPE %s\n", id);

	ptrhash_entry_t *pte = ptrhash_getentry(pipes, (char *) id);
	if(pte) {
		pipe_kill(pte->value, PIR_KILL);
		return 0;
	} else {
		plog(LOG_ERROR, "IGNORING attempt to kill pipe '%s' (it doesn't exist)\n");
		return 0;
	}
}

int p_unpause(void *data, const char *id)
{
	plog(LOG_DEBUG, "Unpausing proc %s\n", id);

	ptrhash_entry_t *pte = ptrhash_getentry(procs, (char *) id);
	if(pte) {
		proc_t *p = pte->value;

		int i;
		for(i=0; i<p->srcs.n; i++) {
			src_t *s = p->srcs.p[i];
			switch(s->type) {
			case SRC_FD:
				{
					src_fd_t *sf = (src_fd_t *) s;
					if(s->live && !sf->read) {
						sf->read = 1;
						src_fd_set_read(sf);
					}
				}
				break;
			default:;
			}
		}

		return 0;
	} else {
		plog(LOG_ERROR, "IGNORING attempt to unpause process '%s' (it doesn't exist)\n", id);
		return 0;
	}
}

int p_timer(void *data, const char *timerid, int repeat, struct timeval delay, const char *output)
{
	plog(LOG_DEBUG, "Setting timer %s\n", timerid);

	ptrhash_entry_t *pte = ptrhash_getentry(timers, (char *) timerid);
	if(pte) {
		plog(LOG_DEBUG, "Killing existing timer %s\n", timerid);
		timerq_t *t = pte->value;

		evtimer_del(&t->ev);

		ptrhash_del(timers, t->id);
		free(t->report);
		free(t->id);
		free(t);
	}

	timerq_t *timerq = malloc(sizeof *timerq);
	if(!timerq) abort();


	timerq->delay = delay;
	timerq->id = strdup(timerid);
	timerq->repeat = repeat;

	/* gotta append endline */
	timerq->reportlen = strlen(output) + 1;
	timerq->report = malloc(timerq->reportlen + 1);
	if(!timerq->report) abort();
	memcpy(timerq->report, output, timerq->reportlen - 1);
	timerq->report[timerq->reportlen-1] = 10;
	timerq->report[timerq->reportlen] = 0;

	evtimer_set(&timerq->ev, timerq_go, timerq);
	evtimer_add(&timerq->ev, &timerq->delay);

	ptrhash_set(timers, timerq->id, timerq);

	return 0;
}

int p_endtimer(void *data, const char *timerid)
{
	plog(LOG_DEBUG, "Ending timer %s\n", timerid);

	ptrhash_entry_t *pte = ptrhash_getentry(timers, (char *) timerid);
	if(pte) {
		timerq_t *t = pte->value;

		evtimer_del(&t->ev);

		ptrhash_del(timers, t->id);
		free(t->report);
		free(t->id);
		free(t);
	} else {
		plog(LOG_ERROR, "IGNORING attempt to delete timer '%s' (it doesn't exist)\n", timerid);
	}

	return 0;
}

/****************************************************************************
 * sources
 ****************************************************************************/

/* initialize common header */
void src_init(src_t *src)
{
	ptrlist_init(&src->pipes);
	src->oneof = term_eof;
	src->live = 1;
	src->postpone = 0;
	src->cascade = 0;
}

src_ev_t *src_ev_new()
{
	src_ev_t *s = malloc(sizeof *s);
	if(!s) abort();
	src_init((src_t *) s);

	s->src.type = SRC_EV;
	s->level = loglevel;

	return s;
}

/* write data to source, plus an endline!
 * clobbers after the buffer!!
 * (to be used with buf_line only, i guess) */
static void src_write_eol(src_t *s, char *line, int len)
{
	line[len] = '\n';
	src_write(s, line, len + 1);
}

src_fd_t *src_fd_new(int fd, int record, termination_t term)
{
	src_fd_t *s = malloc(sizeof *s);
	if(!s) abort();

	src_init((src_t *) s);
	s->src.type = SRC_FD;
	s->src.oneof = term == TERM_DEFAULT ? term_eof : term;

	s->read = 1;
	s->fd = fd;
	s->fdpair[0] = -1;
	s->fdpair[1] = -1;
	s->recordsize = record;
	s->bufat = 0;
	s->direct = 0;
	s->bidir = 0;

	if(s->recordsize < 0) {
		/* line buffer */
		s->buf = buf_line_new(s, (buf_line_call_t) src_write_eol);
	} else if(s->recordsize > 0) {
		/* fixed record buffer */
		s->buf = malloc(s->recordsize);
		if(!s->buf) abort();
	}

	return s;
}

/* set up reading for an FD src */
void src_fd_set_read(src_fd_t *sf)
{
	if(sf->direct) goto direct;

	event_set(&sf->ev, sf->fdpair[0], EV_READ | EV_PERSIST, (void (*) (int, short, void *)) src_fd_read, sf);
	if(!event_add(&sf->ev, NULL)) return;

	sf->direct = 1;
	plog(LOG_DEBUG, "event_set failed for reading fd %d, trying direct reading\n", sf->fdpair[0]);

direct:
	/* use direct reading method
	 * sometimes necessary because epoll doesn't seem to work on file fds
	 */

	evtimer_set(&sf->ev, (void (*) (int, short, void *)) src_fd_read, sf);

	struct timeval tv;
	tv.tv_sec = 0;
	tv.tv_usec = 0;

	if(evtimer_add(&sf->ev, &tv)) abort();
}

/* our global event source */
src_ev_t *global_evsrc = NULL;

/* plumb's global FD srcs */
ptrset_t global_srcs;

/* kill/close source */
void src_end(src_t *s, src_end_reason_t reason)
{
	if(!s->live) {
		plog(LOG_DEBUG, "Can't end already ended source %p\n", s);
		return;
	}

	plog(LOG_DEBUG, "Ending source %p because %s\n", s, src_end_reason_str[reason]);

	s->live = 0;

	switch(s->type) {
	case SRC_FD:
		{
			src_fd_t *sf = (src_fd_t *) s;
			plog(LOG_DEBUG, "Ended source %p is an fd: %d\n", s, sf->fdpair[0]);

			if(sf->read) {
				event_del(&sf->ev);
				sf->read = 0;
			}

			EVUTIL_CLOSESOCKET(sf->fdpair[0]);
			if(sf->recordsize < 0) buf_line_free(sf->buf);
			else if(sf->recordsize > 0) free(sf->buf);
			sf->buf = NULL;
			sf->recordsize = 0;
		}
		break;
	case SRC_EV:
		break;
	}

	pipe_kill_reason_t preason;
	switch(reason) {
	case SER_KILL:
		preason = PIR_KILL;
		break;
	case SER_EOF:
		preason = PIR_EOF;
		break;
	case SER_READ:
		preason = PIR_READ;
		break;
	case SER_CASCADE:
		preason = PIR_CASCADE;
		break;
	};

	/* kill pipes from this source */
	while(s->pipes.first) {
		pipe_kill(s->pipes.first->ptr, preason);
	}

	ptrlist_free(&s->pipes);

	if(s->oneof == TERM_KILL) {
		event_loopbreak();
	}

	/* if the src is postponed, free it */
	if(s->postpone) {
		free(s);
		postponed_srcs --;
	}
}

/* (libevent_callback) there's stuff to read */
void src_fd_read(int fd, short event, src_fd_t *s)
{
	/* src_fd_read might be called from a timer (for direct reading) */
	fd = s->fdpair[0];

	char buf[4096];

	if(s->recordsize < 1) {
		/* line-based or read as possible */
		int r = os_net_read(fd, buf, 4096);
		plog(LOG_JUNK, "Read from fd %d, result %d\n", fd, r);
		if(r < 1) {
			if(r < 0) {
				if(errno == EINTR) return;
				perror("read() error");
				plog(LOG_ERROR, "source read() fails\n");
				src_end((src_t *) s, SER_READ);
			} else {
				src_end((src_t *) s, SER_EOF);
			}

			return;
		}

		if(s->buf == NULL) {
			src_write((src_t *) s, buf, r);
		} else {
			buf_line_write(s->buf, buf, r);
		}
	} else {
		/* constant buffer */
		int r = os_net_read(fd, s->buf, s->recordsize - s->bufat);
		plog(LOG_JUNK, "Read from fd %d, result %d\n", fd, r);
		if(r < 1) {
			if(r < 0) {
				if(errno == EINTR) return;
				perror("read() error");
				plog(LOG_ERROR, "source read() fails\n");
				src_end((src_t *) s, SER_READ);
			} else {
				src_end((src_t *) s, SER_EOF);
			}

			return;
		}

		s->bufat += r;
		if(s->bufat == s->recordsize) {
			src_write((src_t *) s, s->buf, s->recordsize);
			s->bufat = 0;
		}
	}

	/* for direct reading, we need to add another timeout */
	if(s->direct) src_fd_set_read(s);
}

/* write data to source */
void src_write(src_t *s, char *line, int len)
{
	if(!s->live) return;

	ptrlist_el_t *el;

	if(s->type == SRC_FD) {
		src_fd_t *sf = (src_fd_t *) s;
		if(sf->recordsize == -1) {
			/* if it's a line-buffered source, count the lines */
			sf->lines ++;
		}
	}

	/* enumerate pipes from source, write data to all pipes */
	for(el = s->pipes.first; el; el=el->next) {
		pipe_t *p = el->ptr;

		char *data = line;
		int datalen = len;

		/* apply pipe filter */
		if(p->filter != FILT_NONE) {
			if(p->filterlen > len) continue;
			if(memcmp(line, p->filterdata, p->filterlen)) continue;
			if(p->filter == FILT_CUT) {
				data += p->filterlen;
				datalen -= p->filterlen;
			}
		}

		if(p->prefixlen)
			tgt_write(s, p->tgt, p->prefixdata, p->prefixlen);

		if(datalen)
			tgt_write(s, p->tgt, data, datalen);

		if(p->endchain) return;
	}
}

/****************************************************************************
 * targets
 ****************************************************************************/

/* set pipe's number in tgt list */
static void pipe_set_tgtnum(pipe_t *p, int at)
{
	p->tgt_pipe = at;
}

/* initialize common header */
void tgt_init(tgt_t *tgt)
{
	ptrset_init(&tgt->pipes, (ptrset_renum_call_t) pipe_set_tgtnum);
	tgt->onwrite = term_write;
	tgt->live = 1;
	tgt->writing = 0;
	tgt->postpone = 0;
	tgt->cascade = 0;
}

tgt_ctrl_t *tgt_ctrl_new()
{
	tgt_ctrl_t *t = malloc(sizeof *t);
	if(!t) abort();
	tgt_init((tgt_t *) t);

	t->tgt.type = TGT_CTRL;

	return t;
}

tgt_fd_t *tgt_fd_new(int fd, termination_t term)
{
	tgt_fd_t *t = malloc(sizeof *t);
	if(!t) abort();

	tgt_init((tgt_t *) t);
	t->tgt.type = TGT_FD;
	t->tgt.onwrite = term == TERM_DEFAULT ? term_write : term;

	t->fd = fd;
	t->fdpair[0] = -1;
	t->fdpair[1] = -1;
	t->buf = buf_fifo_new(FIFO_PACKET_SIZE);
	t->bidir = 0;

	return t;
}

/* plumb's global FD tgts */
ptrset_t global_tgts;

/* our global event target */
tgt_ctrl_t *global_ctrltgt = NULL;

/* return a descrition of a target as a statically allocated string */
static char *tgt_describe(tgt_t *t)
{
	tgt_fd_t *tfd = (tgt_fd_t *) t;
	int i;
	pipe_t *p;

	strasm_reset();
	bufprintf("target[%p ", t);
	strasm_add(bufprintfbuf);
	switch(t->type) {
	case TGT_FD:
		bufprintf("fd=%d (%d %d)", tfd->fd, tfd->fdpair[0], tfd->fdpair[1]);
		strasm_add(bufprintfbuf);
		break;
	case TGT_CTRL:
		strasm_add("ctrl");
		break;
	}

	if(t->live)
		strasm_add(" live");
	if(t->cascade)
		strasm_add(" cascade");
	if(t->postpone)
		strasm_add(" postpone");
	if(t->writing)
		strasm_add(" writing");

	switch(t->onwrite) {
	case TERM_DEFAULT:
		strasm_add(" onwrite=DEFAULT");
		break;
	case TERM_KILL:
		strasm_add(" onwrite=KILL");
		break;
	case TERM_EVENT:
		strasm_add(" onwrite=EVENT");
		break;
	}

	if(t->pipes.n > 0) {
		strasm_add(" pipes:");
	}

	for(i=0; i<t->pipes.n; i++) {
		p = t->pipes.p[i];
		strasm_add(" ");
		strasm_add(p->id);
	}

	strasm_add("]");

	return strasm_buf;
}

/* kill/close tgt */
void tgt_end(tgt_t *t, tgt_end_reason_t reason)
{
	if(!t->live) {
		plog(LOG_DEBUG, "Can't end already ended target %p\n", t);
		return;
	}

	plog(LOG_DEBUG, "Ending %s because %s\n", tgt_describe(t), tgt_end_reason_str[reason]);

	t->live = 0;

	switch(t->type) {
	case TGT_FD:
		{
			tgt_fd_t *tf = (tgt_fd_t *) t;
			os_closesocket(tf->fdpair[1]);
			buf_fifo_free(tf->buf);
			tf->buf = NULL;
		}
		break;
	case TGT_CTRL:
		break;
	}

	pipe_kill_reason_t preason;
	switch(reason) {
	case TER_KILL:
		preason = PIR_KILL;
		break;
	case TER_WRITE:
		preason = PIR_WRITE;
		break;
	case TER_CASCADE:
		preason = PIR_CASCADE;
		break;
	}

	/* kill pipes to this target */
	while(t->pipes.n > 0) {
		pipe_kill(t->pipes.p[0], preason);
	}

	ptrset_free(&t->pipes);

	/* if the tgt is postponed, free it */
	if(t->postpone) {
		free(t);
		postponed_tgts --;
	}
}

/* write data to target */
void tgt_write(src_t *s, tgt_t *t, char *data, int len)
{
	if(!t->live) return;

	switch(t->type) {
	case TGT_FD:
		{
			tgt_fd_t *tf = (tgt_fd_t *) t;
			buf_fifo_write(tf->buf, data, len);
			if(!tf->tgt.writing && buf_fifo_packetsize(tf->buf)) {
				tf->tgt.writing = 1;
				if(event_once(tf->fdpair[1], EV_WRITE, (void (*) (int, short, void *)) tgt_fd_write, tf, NULL)) {
					plog(LOG_JUNK, "event_once EV_WRITE fails for fd %d, trying a blocking write\n", tf->fdpair[1]);
					return tgt_fd_write(tf->fdpair[1], EV_WRITE, tf);
				}
			}
		}
		break;
	case TGT_CTRL:
		{
			/* ehh.. maybe tok_line could get a len argument? */
			char prev = data[len];
			data[len] = 0;
			/* line buffering appends the zero */
			int r = tok_line(tkn, data, parse_line);
			data[len] = prev;
			if(r) {
				if(r < 0) {
					plog(LOG_ERROR, "Tokenization error (%s) in string: %s\n", tok_error(r), data);
				}
				plog(LOG_DEBUG, "Ending source %p because of ctrl parse error\n", s);

				if(s->type == SRC_FD) {
					src_fd_t *sf = (src_fd_t *) s;
					assert(sf->recordsize == -1);
					plog(LOG_ERROR, "Above parse error occurred in line %d\n", sf->lines);
				}

				src_end(s, SER_READ);
			}
		}
		break;
	}
}

/* (libevent callback) there's stuff to write */
void tgt_fd_write(int fd, short event, tgt_fd_t *t)
{
	assert(event & EV_WRITE);

	t->tgt.writing = 0;

	if(!t->buf) {
		/* tgt got killed in the meanwhile */
		return;
	}

	char *buf;
	int s = buf_fifo_getpacket(t->buf, &buf);
	if(!s) return;

	int r = os_net_write(fd, buf, s);
	if(r < 0) {
		if(errno == EINTR) {
			r = 0;
		} else {
			perror("write() error");
			plog(LOG_ERROR, "target write() fails");
			if(t->tgt.onwrite == TERM_KILL) {
				plog(LOG_DEBUG, "ending plumb because of write error");
				event_loopbreak();
			}
			tgt_end((tgt_t *) t, TER_WRITE);
			return;
		}
	}

	plog(LOG_JUNK, "Written %d bytes to fd %d\n", r, fd);

	buf_fifo_advance(t->buf, r);

	if(buf_fifo_packetsize(t->buf)) {
		t->tgt.writing = 1;
		if(event_once(fd, EV_WRITE, (void (*) (int, short, void *)) tgt_fd_write, t, NULL)) {
			plog(LOG_JUNK, "event_once EV_WRITE fails for fd %d, trying a blocking write\n", fd);
			return tgt_fd_write(fd, event, t);
		}
	} else {
		if(t->tgt.postpone) {
			tgt_end((tgt_t *) t, TER_CASCADE);
		}
	}
}

/****************************************************************************
 * main
 ****************************************************************************/

ptrhash_t *pipes = NULL;
ptrhash_t *procs = NULL;
ptrhash_t *timers = NULL;

int nprocs = 0;
int npipes = 0;
int postponed_srcs = 0;
int postponed_tgts = 0;
int exit_value = 0;

/* destroy/free a process */
void proc_kill(proc_t *p, proc_kill_reason_t reason, int exitv)
{
	if(p->killing) return;
	p->killing = 1;

	plog(LOG_DEBUG, "Killing process %s [%d], reason %s, exitv %d\n", p->id, p->pid, proc_kill_reason_str[reason], exitv);

	int evl = bufprintf("ENDPROC %s %s %d\n", p->id, proc_kill_reason_str[reason], exitv);
	src_write((src_t *) global_evsrc, bufprintfbuf, evl);

	os_kill(p);

	int i;
	for(i=0; i<p->argc; i++)
		free(p->argv[i]);
	free(p->argv);
	p->argv = NULL;

	ptrhash_del(procs, p->id);
	free(p->id);
	p->id = NULL;
	nprocs --;

	for(i=0; i<p->srcs.n; i++) {
		if(((src_t *) p->srcs.p[i])->live) {
			plog(LOG_DEBUG, "Postponing live src %p\n", p->srcs.p[i]);
			((src_t *)p->srcs.p[i])->postpone = 1;
			postponed_srcs ++;
		} else {
			src_end(p->srcs.p[i], SER_KILL);
			free(p->srcs.p[i]);
		}
	}
	for(i=0; i<p->tgts.n; i++) {
		if(((tgt_t *)p->tgts.p[i])->writing) {
			plog(LOG_DEBUG, "Postponing writing tgt %p\n", p->tgts.p[i]);
			((tgt_t *)p->tgts.p[i])->postpone = 1;
			postponed_tgts ++;
		} else {
			tgt_end(p->tgts.p[i], TER_KILL);
			free(p->tgts.p[i]);
		}
	}

	ptrset_free(&p->srcs);
	ptrset_free(&p->tgts);

	if(p->onend == TERM_KILL) {
		event_loopbreak();
	}

	free(p);

	if(!nprocs && !npipes) {
		plog(LOG_DEBUG, "No process or pipe left, ending plumb\n");
		event_loopbreak();
		return;
	}
}

/* destroy/free a pipe */
void pipe_kill(pipe_t *p, pipe_kill_reason_t reason)
{
	plog(LOG_DEBUG, "Killing pipe %s, reason %s\n", p->id, pipe_kill_reason_str[reason]);

	int evl = bufprintf("ENDPIPE %s %s\n", p->id, pipe_kill_reason_str[reason]);
	src_write((src_t *) global_evsrc, bufprintfbuf, evl);

	ptrhash_del(pipes, p->id);
	free(p->id);
	p->id = NULL;

	if(p->src != (src_t *) global_evsrc) npipes --;

	ptrlist_del_el(&p->src->pipes, p->src_el);
	p->src_el = NULL;

	ptrset_del_idx(&p->tgt->pipes, p->tgt_pipe);
	p->tgt_pipe = -1;

	free(p->filterdata);
	free(p->prefixdata);

	if(p->src->cascade && !p->src->pipes.first) {
		src_end(p->src, SER_CASCADE);
	}

	if(p->tgt->cascade && !p->tgt->pipes.n) {
		tgt_end(p->tgt, TER_CASCADE);
	}

	free(p);

	if(!nprocs && !npipes) {
		plog(LOG_DEBUG, "No process or pipe left, ending plumb\n");
		event_loopbreak();
	}
}

/* kill all live subprocesses */
void endplumb()
{
	static int plumb_ending = 0;

	if(plumb_ending) return;
	plumb_ending = 1;

	plog(LOG_DEBUG, "Plumb shutdown\n");

	/* kill all processes */
	ptrhash_entry_t *pe = ptrhash_first(procs);
	for(; pe; pe = ptrhash_next(procs, pe)) {
		proc_kill(pe->value, PKR_KILL, 0);
	}

	/* postpone global tgts */
	int i;
	for(i=0; i<global_tgts.n; i++) {
		tgt_t *t = global_tgts.p[i];
		if(t->writing && !t->postpone) {
			plog(LOG_DEBUG, "Postponing writing global tgt %p\n", t);
			t->postpone = 1;
			postponed_tgts ++;
		}
	}
}

/* do final loop (final writes), then exit */
void finalloop(int status)
{
	while(postponed_tgts > 0) {
		if(event_loop(EVLOOP_ONCE)) {
			plog(LOG_ERROR, "event_loop returned with error\n");
			break;
		}
	}

	exit(status);
}

/* (libevent callback) a timer has triggered */
void timerq_go(int fd, short event, void *data)
{
	timerq_t *t = data;

	src_write((src_t *) global_evsrc, t->report, t->reportlen);

	if(t->repeat) {
		evtimer_add(&t->ev, &t->delay);
	} else {
		ptrhash_del(timers, t->id);
		free(t->report);
		free(t->id);
		free(t);
	}
}

void help(char *us)
{
	printf(
			"Usage: %s <options/commands...>\n\n"
			"  -69 <command1> <command2>   start two programs piped together\n"
			"  -v <key> <value>            set variable to single token\n"
			"  -V <key> <t1> <t2...> .     set variable to multiple tokens\n"
			"                                (end with a single period)\n"
			"  -f <filename>               read commands from a file\n"
			"  <command>                   execute commands\n",
		us
	);
}

static void mode69(char *cmd1, char *cmd2)
{
	/* shortcut for 69 mode for two commands */

	p_pipe_src_t cmd_1_src, cmd_2_src;
	p_pipe_tgt_t cmd_1_tgt, cmd_2_tgt;

	/* TODO: windows support for this... */
	char *cmd_1[] = { "/bin/sh", "-c", cmd1 };
	char *cmd_2[] = { "/bin/sh", "-c", cmd2 };

	if(p_proc(NULL, "_69_1", TERM_DEFAULT, 0, 0, NULL, 0, NULL, 3, cmd_1)) {
		fprintf(stderr, "69 mode: failed to start command 1: %s\n", cmd1);
		endplumb();
		finalloop(1);
	}

	if(p_proc(NULL, "_69_2", TERM_DEFAULT, 0, 0, NULL, 0, NULL, 3, cmd_2)) {
		fprintf(stderr, "69 mode: failed to start command 2: %s\n", cmd2);
		endplumb();
		finalloop(1);
	}

	cmd_1_src.type = PSRC_PROC;
	cmd_1_src.proc = "_69_1";
	cmd_1_src.fd = 1;

	cmd_1_tgt.type = PTGT_PROC;
	cmd_1_tgt.proc = "_69_1";
	cmd_1_tgt.fd = 0;

	cmd_2_src.type = PSRC_PROC;
	cmd_2_src.proc = "_69_2";
	cmd_2_src.fd = 1;

	cmd_2_tgt.type = PTGT_PROC;
	cmd_2_tgt.proc = "_69_2";
	cmd_2_tgt.fd = 0;

	if(p_pipe(NULL, "_69_p_1_2", cmd_1_src, cmd_2_tgt, FILT_NONE, 0, NULL, 0, NULL, 0, PPLACE_LAST)) {
		fprintf(stderr, "69 mode: failed to create pipe 1->2\n");
		endplumb();
		finalloop(1);
	}

	if(p_pipe(NULL, "_69_p_2_1", cmd_2_src, cmd_1_tgt, FILT_NONE, 0, NULL, 0, NULL, 0, PPLACE_LAST)) {
		fprintf(stderr, "69 mode: failed to create pipe 2->1\n");
		endplumb();
		finalloop(1);
	}
}

int main(int argc, char *argv[])
{
	int i;
	char *execname;

	if(getenv("DEFAULT_EOF_EVENT"))
		term_eof = TERM_EVENT;
	if(getenv("DEFAULT_WRITE_EVENT"))
		term_write = TERM_EVENT;

	execname = basename(strdup(argv[0]));

	event_init();

	tkn = tok_init();

	tok_setvar(tkn, "argv0", 1, argv);

	/* TODO: set environment in token variable hash */

	global_evsrc = src_ev_new();
	global_ctrltgt = tgt_ctrl_new();

	pipes = strhash_new();
	procs = strhash_new();
	timers = strhash_new();

	/* initialize global srcs and tgts */

	ptrset_init(&global_srcs, NULL);
	ptrset_init(&global_tgts, NULL);

	os_init();

	/* process command line */

	if(!strcmp(execname, "69")) {
		/* plumb was started as 69 */

		if(argc != 3) {
			fprintf(stderr, "69: need two arguments (two commands)\n");
			endplumb();
			finalloop(1);
		}

		mode69(argv[1], argv[2]);
	} else {
		/* plumb was started as plumb (or something else than 69...) */

		for(i=1; i<argc; i++) {
			if(!strcmp(argv[i], "-h")) {
				help(argv[0]);
				return 0;
			}

			if(!strcmp(argv[i], "-d")) {
				loglevel --;
				continue;
			}

			if(!strcmp(argv[i], "-69")) {
				if(i + 2 >= argc) {
					fprintf(stderr, "Missing arguments for -69\n");
					endplumb();
					finalloop(1);
				}

				mode69(argv[i+1], argv[i+2]);
				i += 2;
				continue;
			}

			if(!strcmp(argv[i], "-v")) {
				if(i + 2 >= argc) {
					fprintf(stderr, "Missing arguments for -v\n");
					endplumb();
					finalloop(1);
				}

				tok_setvar(tkn, argv[i+1], 1, argv+i+2);

				i += 2;
				continue;
			}

			if(!strcmp(argv[i], "-V")) {
				if(i + 2 >= argc) {
					fprintf(stderr, "Missing arguments for -V\n");
					endplumb();
					finalloop(1);
				}

				char *key = argv[i+1];
				int val = i+2;

				for(i+=2; i < argc; i++) {
					if(!strcmp(argv[i], ".")) break;
				}

				if(i == argc) {
					fprintf(stderr, "Missing terminator for -V token list\n");
					endplumb();
					finalloop(1);
				}

				tok_setvar(tkn, key, i-val, argv+val);

				continue;
			}

			if(!strcmp(argv[i], "-f")) {
				if(i + 1 >= argc) {
					fprintf(stderr, "Missing argument for -f\n");
					endplumb();
					finalloop(1);
				}

				i ++;

				char *fn = argv[i];
				FILE *fp = fopen(fn, "r");
				if(!fp) {
					plog(LOG_ERROR, "Failed to open file (%s) for reading\n", fn);
					endplumb();
					finalloop(1);
				}

				int line = 0;

				void file_line(void *data, char *str, int len)
				{
					line ++;

					int r = tok_line(tkn, str, parse_line);
					if(r) {
						if(r < 0) {
							plog(LOG_ERROR, "Tokenization error (%s) in file %s, line %d\n", tok_error(r), fn, line);
						} else {
							plog(LOG_ERROR, "Parse error in file %s, line %d\n", fn, line);
						}
						endplumb();
						finalloop(1);
					}
				}

				buf_line_t *bl = buf_line_new(NULL, file_line);
				char buf[4096];
				size_t r;
				while((r = fread(buf, 1, 4096, fp)) > 0) {
					buf_line_write(bl, buf, r);
				}

				fclose(fp);
				buf_line_free(bl);

				continue;
			}

			/* unrecognized command line arguments will be processed as
			 * ctrl language
			 * for this, we'll have to split them by newlines */

			char *ln = strtok(argv[i], "\n");
			int r;

			while(ln) {
				r = tok_line(tkn, ln, parse_line);
				if(r < 0) {
					plog(LOG_ERROR, "Tokenization error (%s) in string: %s\n", tok_error(r), ln);
				}

				if(r) {
					endplumb();
					finalloop(1);
				}

				ln = strtok(NULL, "\n");
			}
		}
	}

	/* add events for all global srcs */
	for(i=0; i<global_srcs.n; i++) {
		src_t *s = global_srcs.p[i];
		switch(s->type) {
		case SRC_FD:
			/* TODO:
			 * need a way to alter termination behavior (oneof) for srcs
			 * (and everthing else)
			 */
			s->oneof = term_eof;
			src_fd_set_read((src_fd_t *) s);
			break;
		default:;
		}
	}

	if(event_dispatch()) {
		plog(LOG_ERROR, "event_dispatch returned with error\n");
	}

	endplumb();
	finalloop(exit_value);
	return 0;
}
