#ifndef _PLUMB_H
#define _PLUMB_H

#include "buf.h"
#include "parse.h"
#include "util.h"

#include <unistd.h>

#include <event.h>
#include <evutil.h>

#define FIFO_PACKET_SIZE 16384

typedef enum {
	LOG_JUNK,
	LOG_DEBUG,
	LOG_WARNING,
	LOG_ERROR
} log_level_t;

/* main log level */
extern log_level_t loglevel;

/* log something */
void plog(log_level_t lev, const char *fmt, ...);

/* return a concatenated string of tokens
 * uses a static buffer */
const char *ptokens(int tc, char *tv[]);

/* kill all live subprocesses */
void endplumb();
/* do final loop (final writes), then exit */
void finalloop(int status);


typedef struct src src_t;
typedef struct tgt tgt_t;
typedef struct pipe pipe_t;
typedef struct proc proc_t;
typedef struct timerq timerq_t;

/* a pipe */
struct pipe {
	char *id; /* pipe ID */

	src_t *src;
	ptrlist_el_t *src_el; /* element in src pipe list */

	tgt_t *tgt;
	int tgt_pipe; /* number of pipe in tgt pipe list */

	filter_type_t filter;
	char *filterdata;
	int filterlen;
	int endchain;

	char *prefixdata;
	int prefixlen;
};

/* a process */
struct proc {
	char *id; /* process ID */

	ptrset_t srcs;
	ptrset_t tgts;

	pid_t pid;

	int argc;
	char **argv;

	termination_t onend;

	int killing; /* process being killed */
};

/* global hash of pipes */
extern ptrhash_t *pipes;
/* global hash of processes */
extern ptrhash_t *procs;
/* number of active processes */
extern int nprocs;
/* number of live pipes */
extern int npipes;
/* number of postponed srcs */
extern int postponed_srcs;
/* number of postponed tgts */
extern int postponed_tgts;

/* a timer */
struct timerq {
	char *id;
	struct timeval delay;
	int repeat;
	struct event ev;

	char *report;
	int reportlen;
};

/* global hash of timers */
extern ptrhash_t *timers;

/* exit value */
extern int exit_value;

typedef enum {
	PKR_KILL=0,
	PKR_CASCADE=1,
	PKR_EXIT=2
} proc_kill_reason_t;

extern char *proc_kill_reason_str[];

/* destroy/free a process */
void proc_kill(proc_t *p, proc_kill_reason_t reason, int exitv);

typedef enum {
	PIR_KILL=0,
	PIR_EOF=1,
	PIR_READ=2,
	PIR_WRITE=3,
	PIR_CASCADE=4
} pipe_kill_reason_t;

extern char *pipe_kill_reason_str[];

/* destroy/free a pipe */
void pipe_kill(pipe_t *p, pipe_kill_reason_t reason);

/* (libevent callback) a timer has triggered */
void timerq_go(int fd, short event, void *data);

/****************************************************************************
 * sources
 ****************************************************************************/

/* source type */
typedef enum {
	SRC_FD,
	SRC_EV
} src_type_t;

/* common header to all sources */
struct src {
	src_type_t type;
	ptrlist_t pipes;
	termination_t oneof;
	int live;    /* initially 1; 0 after closed */
	int cascade; /* should we close src when last pipe is gone */
	int postpone;/* postponed src (doesn't belong to proc, free on end) */
};

/* initialize common header */
void src_init(src_t *src);

/* fd source */
typedef struct {
	src_t src;

	int read;        /* active reading from FD */

	/* use direct reading instead of libevent
	 * needed to work around epoll problems with file fds */
	int direct;

	int fd;          /* intended source fd (e.g. in the new process) */
	int fdpair[2];   /* the pipe: [0] is the source fd */
	int recordsize;  /* -1: line-buffered 0: send immediately >0: fixed buffer */
	int bidir;       /* bi-directional (see plumb_os_unix.h) */

	void *buf;       /* buffer (buf_line_t, NULL, mallocd fixed buffer) */

	union {
		int bufat;   /* position in fixed buffer (for >0 recordsize) */
		int lines;   /* number of read lines (for line buffering) */
	};

	struct event ev; /* read event structure */
} src_fd_t;

/* event source */
typedef struct {
	src_t src;

	log_level_t level;
} src_ev_t;

src_ev_t *src_ev_new();
src_fd_t *src_fd_new(int fd, int record, termination_t term);

/* set up reading for an FD src */
void src_fd_set_read(src_fd_t *sf);

/* our global event source */
extern src_ev_t *global_evsrc;
/* plumb's global FD srcs */
extern ptrset_t global_srcs;

typedef enum {
	SER_KILL=0,
	SER_EOF=1,
	SER_READ=2,
	SER_CASCADE=3
} src_end_reason_t;

extern char *src_end_reason_str[];

/* kill/close src */
void src_end(src_t *s, src_end_reason_t reason);

/* (libevent callback) there's stuff to read */
void src_fd_read(int fd, short event, src_fd_t *s);

/* write data to source */
void src_write(src_t *s, char *line, int len);

/****************************************************************************
 * targets
 ****************************************************************************/

/* target type */
typedef enum {
	TGT_FD,
	TGT_CTRL
} tgt_type_t;

/* common header to all targets */
struct tgt {
	tgt_type_t type;
	ptrset_t pipes;
	termination_t onwrite;
	int live;
	int cascade; /* should we close tgt when last pipe is gone */
	int postpone;/* postponed tgt (doesn't belong to proc, free on end) */
	int writing; /* whether we're waiting for writing */
};

/* initialize common header */
void tgt_init(tgt_t *tgt);

/* fd target */
typedef struct {
	tgt_t tgt;

	int fd;        /* intended target fd (e.g. in the new process) */
	int fdpair[2]; /* the pipe: [1] is the target fd */
	buf_fifo_t *buf;

	int bidir;       /* bi-directional (see plumb_os_unix.h) */
} tgt_fd_t;

/* ctrl target */
typedef struct {
	tgt_t tgt;
} tgt_ctrl_t;

tgt_ctrl_t *tgt_ctrl_new();
tgt_fd_t *tgt_fd_new(int fd, termination_t term);

/* our global event target */
extern tgt_ctrl_t *global_ctrltgt;
/* plumb's global FD tgts */
extern ptrset_t global_tgts;

typedef enum {
	TER_KILL=0,
	TER_WRITE=1,
	TER_CASCADE=2
} tgt_end_reason_t;

extern char *tgt_end_reason_str[];

/* kill/close tgt */
void tgt_end(tgt_t *t, tgt_end_reason_t reason);

/* (libevent callback) there's stuff to write */
void tgt_fd_write(int fd, short event, tgt_fd_t *t);

/* write data to target */
void tgt_write(src_t *s, tgt_t *t, char *data, int len);

#endif
