#include <sys/wait.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <fcntl.h>

#include "osunix.h"

/* handle SIGCHLD */
static void handle_sigchld(int sig, short what, void *arg)
{
	plog(LOG_DEBUG, "Received SIGCHLD\n");

	int status;
	pid_t chld;

	while((chld = waitpid(-1, &status, WNOHANG)) > 0) {
		plog(LOG_DEBUG, "Child %d is dead\n", chld);

		/* TODO: need better way of looking up process by pid */
		ptrhash_entry_t *pe = ptrhash_first(procs);
		proc_t *p;
		for(; pe; pe = ptrhash_next(procs, pe)) {
			p = pe->value;
			if(p->pid == chld) break;
		}
		if(!pe) {
			plog(LOG_WARNING, "Dead child %d unrecorded!\n", chld);
		} else {
			if(WIFEXITED(status)) {
				proc_kill(p, PKR_EXIT, WEXITSTATUS(status));
			} else {
				plog(LOG_DEBUG, "Child %d exited abnormally\n", chld);
				proc_kill(p, PKR_EXIT, -666);
			}
		}
	}
}

static struct event evchld;

static void os_init()
{
	/* listen for SIGCHLD */
	signal_set(&evchld, SIGCHLD, handle_sigchld, NULL);
	signal_add(&evchld, NULL);

	/* don't keep stdin/stdout in forked processes
	 * stderr is needed for error reporting... */
	fcntl(0, F_SETFD, fcntl(0, F_GETFD) | FD_CLOEXEC);
	fcntl(1, F_SETFD, fcntl(1, F_GETFD) | FD_CLOEXEC);

	/* plumb's stdin */
	src_fd_t *s = src_fd_new(0, -1, term_eof);
	s->src.cascade = 0;
	s->fdpair[0] = 0;
	ptrset_add(&global_srcs, s);

	/* plumb's stdout */
	tgt_fd_t *t = tgt_fd_new(1, term_write);
	t->tgt.cascade = 0;
	t->fdpair[1] = 1;
	ptrset_add(&global_tgts, t);

	/* plumb's stderr */
	t = tgt_fd_new(2, term_write);
	t->tgt.cascade = 0;
	t->fdpair[1] = 2;
	ptrset_add(&global_tgts, t);
}

static int os_fork(proc_t *p, int pause)
{
	int i, j;

	/* create pipes */

	for(i=0; i<p->srcs.n; i++) {
		src_t *s = p->srcs.p[i];
		assert(s->type == SRC_FD);
		src_fd_t *sf = (src_fd_t *) s;
		tgt_fd_t *tbi = NULL;

		for(j=0; j<p->tgts.n; j++) {
			tgt_t *t = p->tgts.p[j];
			assert(t->type == TGT_FD);
			tgt_fd_t *tf = (tgt_fd_t *) t;

			if(sf->fd == tf->fd) {
				sf->bidir = 1;
				tf->bidir = 1;
				tbi = tf;
			}
		}

		if(sf->bidir) {
			plog(LOG_DEBUG, "Process source %p is bidirectional: fd %d with target %p\n", sf, sf->fd, tbi);
			int r = socketpair(AF_UNIX, SOCK_STREAM, 0, sf->fdpair);
			if(r) {
				perror("plumb fatal error: socketpair() fails");
				endplumb();
				exit(1);
			}

			tbi->fdpair[1] = sf->fdpair[0];
			tbi->fdpair[0] = sf->fdpair[1];
		} else {
			int r = pipe(sf->fdpair);
			if(r) {
				perror("plumb fatal error: pipe() fails");
				endplumb();
				exit(1);
			}
		}
	}

	for(i=0; i<p->tgts.n; i++) {
		tgt_t *t = p->tgts.p[i];
		assert(t->type == TGT_FD);
		tgt_fd_t *tf = (tgt_fd_t *) t;

		if(tf->bidir) continue;

		int r = pipe(tf->fdpair);
		if(r) {
			perror("plumb fatal error: pipe() fails");
			endplumb();
			exit(1);
		}
	}

	/* fork off new process */
	p->pid = fork();
	if(p->pid == -1) {
		perror("plumb fatal error: fork() fails");
		endplumb();
		exit(1);
	}

	if(p->pid == 0) {
		/* child side */

		/* process sources
		 *  close unused read end
		 *  assign write end to proper fd
		 */
		for(i=0; i<p->srcs.n; i++) {
			src_t *s = p->srcs.p[i];
			assert(s->type == SRC_FD);
			src_fd_t *sf = (src_fd_t *) s;

			close(sf->fdpair[0]);
			dup2(sf->fdpair[1], sf->fd);
			close(sf->fdpair[1]);
		}

		/* process targets
		 *  close unused write end
		 *  assign read end to proper fd
		 */
		for(i=0; i<p->tgts.n; i++) {
			tgt_t *t = p->tgts.p[i];
			assert(t->type == TGT_FD);
			tgt_fd_t *tf = (tgt_fd_t *) t;

			if(tf->bidir) continue;

			close(tf->fdpair[1]);
			dup2(tf->fdpair[0], tf->fd);
			close(tf->fdpair[0]);
		}

		execvp(p->argv[0], p->argv);
		fprintf(stderr, "Failed to execute process: %s\n", tok_print(p->argc, p->argv));
		exit(1);
	} else {
		/* parent side */

		/* process sources
		 *  close unused write end, set flags on read end
		 */
		for(i=0; i<p->srcs.n; i++) {
			src_t *s = p->srcs.p[i];
			assert(s->type == SRC_FD);
			src_fd_t *sf = (src_fd_t *) s;

			close(sf->fdpair[1]);
			sf->fdpair[1] = -1;

			/* XXX: hello.. isn't this a race condition?
			 * what's FD_CLOEXEC doing here? */
			fcntl(sf->fdpair[0], F_SETFD, fcntl(sf->fdpair[0], F_GETFD) | FD_CLOEXEC);

			plog(LOG_DEBUG, "PROC %s reading from fd %d\n", p->id, sf->fdpair[0]);

			if(pause) {
				sf->read = 0;
			} else {
				assert(sf->read);
				src_fd_set_read(sf);
			}
		}

		/* process targets
		 *  close unused read end, set flags on write end
		 */
		for(i=0; i<p->tgts.n; i++) {
			tgt_t *t = p->tgts.p[i];
			assert(t->type == TGT_FD);
			tgt_fd_t *tf = (tgt_fd_t *) t;

			close(tf->fdpair[0]);
			tf->fdpair[0] = -1;
			fcntl(tf->fdpair[1], F_SETFD, fcntl(tf->fdpair[1], F_GETFD) | FD_CLOEXEC);

			plog(LOG_DEBUG, "PROC %s writing to fd %d\n", p->id, tf->fdpair[1]);
		}

		plog(LOG_DEBUG, "PROC %s pid %d\n", p->id, p->pid);
	}

	return 0;
}

static void os_kill(proc_t *p)
{
	if(p->pid != -1) {
		if(kill(p->pid, SIGTERM)) {
			plog(LOG_DEBUG, "Kill process %s [%d] failed because %s\n", p->id, p->pid, strerror(errno));
		}
		p->pid = -1;
	}
}
