#include <windows.h>
#include <winsock2.h>

#include "oswin32.h"

/* value written in (proc_t *)->killing by watcher thread */
#define PROCDEADMARKER 256

static struct event evchld;
static int evchld_read, evchld_write;

/* can read from the evchld pipe - something happened with a child process */
static void os_evchld_read(int fd, short event, void *data)
{
	char b;
	int r;

	r = recv(fd, &b, 1, 0);
	if(r < 0) {
		fprintf(stderr, "os_evchld_read: recv fails\n");
		return;
	}

	if(r < 1) return;

	/* look for a dead process */
	ptrhash_entry_t *pe = ptrhash_first(procs);
	proc_t *p;
	for(; pe; pe = ptrhash_next(procs, pe)) {
		p = pe->value;
		if(p->killing == PROCDEADMARKER) break;
	}
	if(!pe) {
		plog(LOG_WARNING, "Dead child not found!\n");
	} else {
		DWORD status;
		GetExitCodeProcess((HANDLE) p->pid, &status);
		p->killing = 0;
		proc_kill(p, PKR_EXIT, status);
	}
}

/* thread: waits until a process dies */
static DWORD os_wait_process(LPVOID data)
{
	proc_t *p = data;

	DWORD r = WaitForSingleObject((HANDLE) p->pid, INFINITE);
	if(r == WAIT_FAILED) {
		fprintf(stderr, "os_wait_process: WaitForSingleObject fails\n");
	}

	p->killing = PROCDEADMARKER;

	char b = 1;
	if(send(evchld_write, &b, 1, 0) != 1) {
		fprintf(stderr, "os_wait_process: send evchld_write fails\n");
	}

	return 0;
}

static void os_init()
{
	int r;
	WSADATA wsaData;

	r = WSAStartup(MAKEWORD(2,0), &wsaData);
	if(r != 0) {
		fprintf(stderr, "WSAStartup failed\n");
		exit(1);
	}

	/* make a pipe for sending evchld notification bytes */
	int pipe[2];
	r = evutil_socketpair(AF_UNIX, SOCK_STREAM, 0, pipe);
	if(r) {
		fprintf(stderr, "os_init: evutil_socketpair fails: %d\n", r);
		exit(1);
	}

	evchld_read = pipe[0];
	evchld_write = pipe[1];

	event_set(&evchld, evchld_read, EV_READ|EV_PERSIST, os_evchld_read, NULL);
	if(event_add(&evchld, NULL)) {
		fprintf(stderr, "os_init: event_set fails on evchld\n");
		exit(1);
	}

	/* plumb's stdin */
	src_fd_t *s = src_fd_new(0, -1, term_eof);
	s->src.cascade = 0;
	s->fdpair[0] = os_wrap_read((int) GetStdHandle(STD_INPUT_HANDLE));
	ptrset_add(&global_srcs, s);

	/* plumb's stdout */
	tgt_fd_t *t = tgt_fd_new(1, term_write);
	t->tgt.cascade = 0;
	t->fdpair[1] = os_wrap_write((int) GetStdHandle(STD_OUTPUT_HANDLE));
	ptrset_add(&global_tgts, t);

	/* plumb's stderr */
	t = tgt_fd_new(2, term_write);
	t->tgt.cascade = 0;
	t->fdpair[1] = os_wrap_write((int) GetStdHandle(STD_ERROR_HANDLE));
	ptrset_add(&global_tgts, t);
}

static int os_fork(proc_t *p, int pause)
{
	int i, j;
	char *cmdline;
	char *sp;

	/* calculate an upper bound for cmdline length */
	j = 1;
	for(i=0; i<p->argc; i++) {
		j += 3;
		j += strlen(p->argv[i]) * 2;
	}

	cmdline = malloc(j);
	if(!cmdline) abort();

	/* generate cmdline */
	j=0;
	for(i=0; i<p->argc; i++) {
		if(i) cmdline[j++] = ' ';

		cmdline[j++] = '"';

		for(sp = p->argv[i]; *sp; sp++) {
			switch(*sp) {
			case '"':
				cmdline[j++] = '\\';
				cmdline[j++] = '"';
				break;
			case '\\':
				cmdline[j++] = '\\';
				cmdline[j++] = '\\';
				break;
			default:
				cmdline[j++] = *sp;
			}
		}

		cmdline[j++] = '"';
	}

	cmdline[j] = 0;

	plog(LOG_DEBUG, "os_fork %d = %s\n", p->id, cmdline);

	/* make three pipes for stdin, stdout, stderr */

	int pipe_read[3];
	int pipe_write[3];
	int pipe_dup[3];
	int pipe_sockets[3];

	for(i=0; i<3; i++) {
		if(!CreatePipe((HANDLE *) &pipe_read[i], (HANDLE *) &pipe_write[i], NULL, 0)) {
			plog(LOG_ERROR, "os_fork CreatePipe failed\n");
			endplumb();
			exit(1);
		}
	}

	DuplicateHandle(GetCurrentProcess(), (HANDLE) pipe_read[0], GetCurrentProcess(), (HANDLE *) &pipe_dup[0], 0, TRUE, DUPLICATE_SAME_ACCESS | DUPLICATE_CLOSE_SOURCE);
	DuplicateHandle(GetCurrentProcess(), (HANDLE) pipe_write[1], GetCurrentProcess(), (HANDLE *) &pipe_dup[1], 0, TRUE, DUPLICATE_SAME_ACCESS | DUPLICATE_CLOSE_SOURCE);
	DuplicateHandle(GetCurrentProcess(), (HANDLE) pipe_write[2], GetCurrentProcess(), (HANDLE *) &pipe_dup[2], 0, TRUE, DUPLICATE_SAME_ACCESS | DUPLICATE_CLOSE_SOURCE);

	/* create new process */

	PROCESS_INFORMATION procinfo;
	STARTUPINFO startinfo;

	ZeroMemory(&procinfo, sizeof(PROCESS_INFORMATION));
	ZeroMemory(&startinfo, sizeof(STARTUPINFO));

	startinfo.cb = sizeof(STARTUPINFO);
	startinfo.hStdInput = (HANDLE) pipe_dup[0];
	startinfo.hStdOutput = (HANDLE) pipe_dup[1];
	startinfo.hStdError = (HANDLE) pipe_dup[2];
	startinfo.dwFlags |= STARTF_USESTDHANDLES;

	if(CreateProcess(NULL,
		cmdline,
		NULL, /* security attribs */
		NULL, /* primary thread security attribs */
		TRUE, /* handles are inherited */
		0,    /* creation flags */
		NULL, /* use parent's environment */
		NULL, /* use parent's current directory */
		&startinfo,
		&procinfo)
	== 0) {
		/* failed to create process... */
		for(i=0; i<3; i++) {
			CloseHandle((HANDLE) pipe_read[i]);
			CloseHandle((HANDLE) pipe_write[i]);
		}

		plog(LOG_DEBUG, "failed to CreateProcess %s, because %d\n", p->id, GetLastError());
		return 1;
	}

	p->pid = (int) procinfo.hProcess;
	CloseHandle(procinfo.hThread);

	for(i=0; i<3; i++)
		CloseHandle((HANDLE) pipe_dup[i]);

	//fprintf(stderr, "sockets: %d %d %d\n", pipe_write[0], pipe_read[1], pipe_read[2]);

	pipe_sockets[0] = os_wrap_write(pipe_write[0]);
	pipe_sockets[1] = os_wrap_read(pipe_read[1]);
	pipe_sockets[2] = os_wrap_read(pipe_read[2]);

	/* assign source/target pipes */

	for(i=0; i<p->srcs.n; i++) {
		src_t *s = p->srcs.p[i];
		assert(s->type == SRC_FD);
		src_fd_t *sf = (src_fd_t *) s;

		if(sf->fd == 1) {
			sf->fdpair[0] = pipe_sockets[1];
		} else if(sf->fd == 2) {
			sf->fdpair[0] = pipe_sockets[2];
		} else {
			plog(LOG_ERROR, "win32 plumb doesn't support non-1,2 source fds\n");
			endplumb();
			exit(1);
		}

		plog(LOG_DEBUG, "PROC %s reading from fd %d\n", p->id, sf->fdpair[0]);

		if(pause) {
			sf->read = 0;
		} else {
			assert(sf->read);
			src_fd_set_read(sf);
		}
	}

	for(i=0; i<p->tgts.n; i++) {
		tgt_t *t = p->tgts.p[i];
		assert(t->type == TGT_FD);
		tgt_fd_t *tf = (tgt_fd_t *) t;

		if(tf->fd == 0) {
			tf->fdpair[1] = pipe_sockets[0];
		} else {
			plog(LOG_ERROR, "win32 plumb doesn't support non-0 target fds\n");
			endplumb();
			exit(1);
		}

		plog(LOG_DEBUG, "PROC %s writing to fd %d\n", p->id, tf->fdpair[1]);
	}

	plog(LOG_DEBUG, "PROC %s pid %d\n", p->id, p->pid);

	HANDLE t = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE) os_wait_process, p, 0, NULL);
	if(t == NULL) {
		plog(LOG_ERROR, "failed to create waiter thread for proc %s\n", p->id);
		endplumb();
		exit(1);
	}

	return 0;
}

static void os_kill(proc_t *p)
{
	if(p->pid != -1) {
		HANDLE ph = (HANDLE) p->pid;
		DWORD exitCode = 0;

		if(!GetExitCodeProcess(ph, &exitCode)) {
			plog(LOG_DEBUG, "GetExitCodeProcess %s [%d] failed: %d\n", p->id, p->pid, (int) GetLastError());
		}

		if(!TerminateProcess(ph, exitCode)) {
			plog(LOG_DEBUG, "Kill process %s [%d] failed: %d, exitCode=%lx\n", p->id, p->pid, (int) GetLastError(), exitCode);
		}

		CloseHandle(ph);

		p->pid = -1;
	}
}
