#include "buf.h"
#include "util.h"

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <stdarg.h>
#include <string.h>

#include <event.h>

#ifdef WIN32
#include "plumb_net_win32.h"
#else
#include "plumb_net_unix.h"
#endif

/* whether to enable TCP_NODELAY */
int nodelay = 0;

/* whether to report EOF instead of END */
int reporteof = 0;

/* connection info */
typedef struct {
	char id[16];
	char prefix[16];
	int prefixlen;
	int fd;
	int end;
	struct bufferevent *bev;
} conn_t;

/* connection ID counter */
int nextconnid = 0;

/* this hash stores connections */
ptrhash_t *conns;

/* bufferevent for stdin */
struct bufferevent *stdinbev;

/* read from connection */
static void conn_read(struct bufferevent *bufev, conn_t *conn)
{
	char *line;
	int len;

	struct evbuffer *rb = bufev->input;
	if(!rb) return;

next:
	line = evbuffer_readline(rb);
	if(line == NULL) return;

	len = strlen(line);

	line[len] = '\n';
	printf("%s ", conn->id);
	fwrite(line, len+1, 1, stdout);
	fflush(stdout);

	free(line);
	goto next;
}

/* kill a connection */
static void conn_kill(conn_t *conn)
{
	printf("END %s\n", conn->id);
	fflush(stdout);

	bufferevent_free(conn->bev);
	os_net_close(conn->fd);
	ptrhash_del(conns, conn->id);
	free(conn);
}

/* connection write callback (used only when need to kill!) */
static void conn_write(struct bufferevent *bufev, conn_t *conn)
{
	if(conn->end) conn_kill(conn);
}

/* error on connection */
static void conn_error(struct bufferevent *bufev, short what, conn_t *conn)
{
	if(!(what & EVBUFFER_EOF)) {
		fprintf(stderr, "Error on connection %s\n", conn->id);
		conn_kill(conn);
		return;
	}

	bufferevent_disable(conn->bev, EV_READ);

	if(reporteof) {
		printf("EOF %s\n", conn->id);
		fflush(stdout);
	} else {
		bufferevent_enable(conn->bev, EV_WRITE);
		conn->end = 1;
	}
}

/* handle new connection */
static void newconn(int fd, short event, void *arg)
{
	conn_t *conn;
	struct sockaddr_in cli_addr;

	socket_t newsockfd;
	if(os_net_accept(fd, &newsockfd, &cli_addr)) {
		fprintf(stderr, "accept failed\n");
		return;
	}

	if(nodelay) os_net_nodelay(newsockfd);

	conn = malloc(sizeof *conn);
	if(!conn) abort();

	nextconnid ++;
	sprintf(conn->id, "%d", nextconnid);
	conn->prefixlen = sprintf(conn->prefix, "%s ", conn->id);

	conn->end = 0;
	conn->fd = newsockfd;
	conn->bev = bufferevent_new(newsockfd, (evbuffercb) conn_read, (evbuffercb) conn_write, (everrorcb) conn_error, conn);
	bufferevent_enable(conn->bev, EV_READ);

	printf("NEW %s %s\n", conn->id, inet_ntoa(cli_addr.sin_addr));
	fflush(stdout);

	ptrhash_set(conns, conn->id, conn);
}

/* read a line from stdin */
static void stdin_read_gotline(void *cbdata, char *line, int llen)
{
	char *p, *data;
	ptrhash_entry_t *pe;
	conn_t *conn;
	int len;

	p = strchr(line, ' ');
	if(!p) {
		fprintf(stderr, "Syntax error on STDIN: no space in line\n");
		exit(1);
	}

	*p = 0;

	data = p + 1;
	len = strlen(data);

	if(!strcmp(line, "*")) {
		/* broadcast */
		data[len] = '\n';
		pe = ptrhash_first(conns);
		for(; pe; pe = ptrhash_next(conns, pe)) {
			conn = pe->value;
			if(bufferevent_write(conn->bev, data, len+1)) {
				fprintf(stderr, "bufferevent_write to conn %s failed\n", conn->id);
			}
		}
	} else if(!strcmp(line, "end")) {
		/* end a connection */
		pe = ptrhash_getentry(conns, data);
		if(!pe) {
			fprintf(stderr, "end %s: conn with id %s not found\n", data, data);
		} else {
			conn = pe->value;
			bufferevent_disable(conn->bev, EV_READ);
			bufferevent_enable(conn->bev, EV_WRITE);
			conn->end = 1;
		}
	} else if(!strcmp(line, "kill")) {
		/* kill a connection */
		pe = ptrhash_getentry(conns, data);
		if(!pe) {
			fprintf(stderr, "kill %s: conn with id %s not found\n", data, data);
		} else {
			conn_kill(pe->value);
		}
	} else {
		/* ordinary transfer */
		pe = ptrhash_getentry(conns, line);
		if(!pe) {
			fprintf(stderr, "trying to write line to non-existent connection %s\n", line);
		} else {
			conn = pe->value;
			data[len] = '\n';
			if(bufferevent_write(conn->bev, data, len+1)) {
				fprintf(stderr, "bufferevent_write to conn %s failed\n", conn->id);
			}
		}
	}
}

static struct event ev_stdin, ev_sock;
static buf_line_t *stdin_buf;
static int nc_read_shutdown = 1;

/* read from stdin */
static void stdin_read(int fd, short event, void *arg)
{
	char buf[4096];
	int r;

	r = os_net_read(fd, buf, 4096);

	if(r < 1) {
		if(r < 0) {
#ifndef WIN32
			if(errno == EINTR) return;
#endif
			perror("read() error on stdin");
			os_net_close(fd);
			event_del(&ev_stdin);
		} else {
			/* EOF on stdin */
			os_net_close(fd);
			event_del(&ev_stdin);
			event_loopbreak();
		}

		return;
	}

	buf_line_write(stdin_buf, buf, r);
}

static int nc_live = 2;

/* netcat mode: read from stdin */
static void nc_stdin_read(int fd, short event, void *arg)
{
	int sock = *((int *) arg);
	char buf[2048];

	int r = os_net_read(fd, buf, 2048);
	int at = 0;
	int w;

	if(r <= 0) {
		if(r < 0) perror("STDIN read error");

		os_net_close(fd);
		shutdown(sock, SHUT_WR);
		event_del(&ev_stdin);
		nc_live --;
		if(!nc_live) event_loopbreak();
		return;
	}

	while(at < r) {
		w = send(sock, buf+at, r-at, 0);
		if(w < 0) {
			perror("socket write error");
			os_net_close(fd);
			shutdown(sock, SHUT_WR);
			event_del(&ev_stdin);
			nc_live --;
			if(!nc_live) event_loopbreak();
			return;
		}
		at += w;
	}
}

/* netcat mode: read all of stdin
 * returns 1 on error */
int read_all_stdin(int fd, int sock)
{
	char buf[2048];
	int r;
	int at;
	int w;

	for(;;) {
		r = os_net_read(fd, buf, 2048);
		at = 0;

		if(r <= 0) {
			if(r < 0) perror("STDIN read error");

			os_net_close(fd);
			shutdown(sock, SHUT_WR);
			return 0;
		}

		while(at < r) {
			w = send(sock, buf+at, r-at, 0);
			if(w < 0) {
				perror("socket write error");
				os_net_close(fd);
				shutdown(sock, SHUT_WR);
				return 1;
			}
			at += w;
		}
	}
}

/* netcat mode: read from socket */
static void nc_sock_read(int fd, short event, void *arg)
{
	char buf[2048];

	int r = recv(fd, buf, 2048, 0);
	int at = 0;
	int w;

	if(r <= 0) {
		if(r < 0) perror("socket read error");

		shutdown(fd, SHUT_RD);
		close(1);
		event_del(&ev_sock);
		nc_live --;
		if(!nc_live) event_loopbreak();
		else if(nc_read_shutdown) {
			shutdown(fd, SHUT_WR);
			os_net_close(fd);
			event_loopbreak();
		}

		return;
	}

	while(at < r) {
		w = write(1, buf+at, r-at);
		if(w < 0) {
			perror("STDOUT write error");
			close(1);
			shutdown(fd, SHUT_RD);
			event_del(&ev_sock);
			nc_live --;
			if(!nc_live) event_loopbreak();
			return;
		}
		at += w;
	}
}

static void help(char *prog)
{
	printf("Usage: %s [-i iface] -p port [-n] [-c hostname] [-s]\n", prog);
	printf("  -i iface      sets the interface to bind to (default 127.0.0.1)\n");
	printf("  -p port       sets the port\n");
	printf("  -n            enables TCP_NODELAY\n");
	printf("  -c hostname   netcat mode: connect to hostname (on the given -p port)\n");
	printf("  -s            in netcat mode: shutdown only after EOF both on STDIN and socket\n");
	printf("                in plumbnet mode: report EOF instead of END on socket EOF\n");
}

int main(int argc, char *argv[])
{
	char *iface = "127.0.0.1";
	char *host = NULL;
	int port = -1;

	int opt;

	int sock, infd;

	struct event ev1;

	while((opt = getopt(argc, argv, "c:i:p:nhs")) != -1) {
		switch(opt) {
		case 'c':
			host = optarg;
			break;
		case 'i':
			iface = optarg;
			break;
		case 'p':
			port = atoi(optarg);
			break;
		case 'h':
			help(argv[0]);
			exit(0);
		case 'n':
			nodelay = 1;
			break;
		case 's':
			nc_read_shutdown = 0;
			reporteof = 1;
			break;
		default:
			fprintf(stderr, "Invalid option, see %s -h for help\n", argv[0]);
			exit(1);
		}
	}

	if(optind < argc) {
		fprintf(stderr, "Unexpected arguments after options\n");
		exit(1);
	}

	if(port < 0 || port > 65535) {
		fprintf(stderr, "Please specify a port (see %s -h)\n", argv[0]);
		exit(1);
	}

	os_net_init();

	event_init();

	if(host) {
		/* TCP client (netcat) mode */

		sock = os_net_client(host, port);

		event_set(&ev_sock, sock, EV_READ|EV_PERSIST, nc_sock_read, NULL);
		if(event_add(&ev_sock, NULL)) {
			fprintf(stderr, "event_add sock failed\n");
			exit(1);
		}

		infd = os_stdin_fd();
		event_set(&ev_stdin, infd, EV_READ|EV_PERSIST, nc_stdin_read, &sock);
		if(event_add(&ev_stdin, NULL)) {
			/* in this case, stdin is probably an ordinary file,
			 * try to read all and send it to the socket */

			if(read_all_stdin(infd, sock)) exit(1);
		}
	} else {
		/* TCP server mode */

		/* set up TCP server */
		sock = os_net_server(port, iface);

		/* set up main loop */
		conns = strhash_new();

		event_set(&ev1, sock, EV_READ|EV_PERSIST, newconn, NULL);
		if(event_add(&ev1, NULL)) {
			fprintf(stderr, "event_add sock failed\n");
			exit(1);
		}

		stdin_buf = buf_line_new(NULL, stdin_read_gotline);

		event_set(&ev_stdin, os_stdin_fd(), EV_READ|EV_PERSIST, stdin_read, NULL);
		if(event_add(&ev_stdin, NULL)) {
			fprintf(stderr, "event_add stdin failed\n");
			exit(1);
		}
	}

	if(event_dispatch()) {
		fprintf(stderr, "event_dispatch returned error\n");
		exit(1);
	}

	return 0;
}
