#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>

#include "tokens.h"
#include "util.h"

struct tok_state {
	ptrhash_t * vars; /* string->var_t* hash */

	/* token buffer */
	char **tokv;
	int tokalloc;
	int tokc;
	/* string buffer */
	char *sbuf;
	int salloc;
};

/* variable record */
typedef struct {
	int tokc;
	char *tokv[1];
} var_t;

/* free variable record */
static void var_free(var_t *v)
{
	int i;

	for(i=0; i<v->tokc; i++) {
		free(v->tokv[i]);
	}

	free(v);
}

/* create variable record */
static var_t *var_make(char **tokv, int tokc)
{
	int i;
	var_t *v = malloc(sizeof(var_t) + sizeof(char *) * tokc);

	v->tokc = tokc;

	for(i=0; i<tokc; i++) {
		v->tokv[i] = strdup(tokv[i]);
	}

	return v;
}

/* create tokenizer */
tok_t * tok_init()
{
	tok_t *t = malloc(sizeof *t);
	if(!t) abort();

	t->vars = strhash_new();
	t->tokv = NULL;
	t->tokalloc = 0;
	t->tokc = 0;
	t->sbuf = NULL;
	t->salloc = 0;

	return t;
}

/* destroy tokenizer */
void tok_free(tok_t *t)
{
	ptrhash_entry_t *pe = ptrhash_first(t->vars);
	for(; pe; pe = ptrhash_next(t->vars, pe)) {
		free(pe->key);
		var_free(pe->value);
	}

	ptrhash_free(t->vars);

	free(t->tokv);
	free(t->sbuf);

	free(t);
}

static char * skipspace(char *p)
{
	while(*p && isspace(*p)) p++;
	return p;
}

static char * skipuntilspace(char *p)
{
	while(*p && !isspace(*p) && *p != '#' && *p != '"' && *p != ';') p++;
	return p;
}

/* close string at p, advance p
 * (unless it would run out of the buffer) */
static void tclose(int *nextch, char **p)
{
	*nextch = -1;
	if(**p) {
		if(!isspace(**p)) *nextch = **p;
		**p = 0;
		(*p) ++;
	}
}

static void addtoken(tok_t *t, char *tk)
{
	if(t->tokc >= t->tokalloc) {
		t->tokalloc += 8;
		t->tokv = realloc(t->tokv, t->tokalloc * sizeof(t->tokv[0]));
		if(!t->tokv) abort();
	}
	t->tokv[t->tokc++] = tk;
}

static int unhex(char c)
{
	if(c >= '0' && c <= '9') return c - '0';
	c = toupper(c);
	if(c >= 'A' && c <= 'F') return c - 'A' + 10;
	return -1;
}

/* process a line
 * callback will be called with each tokenized command
 * will return one of the tok_result enum (0 on success),
 * or the return value of the first nonzero callback */
int tok_line(tok_t *t, const char *line, tok_line_call_t cb)
{
	int nextch=-1;
	int ll = strlen(line) + 1;
	char *p, *q, *ref;
	char **vtokv;
	int ch, i, vtokc, minor, major;

	if(t->salloc < ll) {
		t->sbuf = realloc(t->sbuf, ll);
		if(!t->sbuf) abort();
		t->salloc = ll;
	}
	strcpy(t->sbuf, line);

	p = t->sbuf;

	for(;;) {
		/* beginning of a new sentence */

		if(!*p) break;

		t->tokc = 0;

		for(;;) {
			/* beginning of a new token */

			if(nextch == -1) {
				p = skipspace(p);
				if(!*p) break;
				ch = p[0];
			} else {
				p --;
				ch = nextch;
				nextch = -1;
			}

			if(ch == '#') {
				/* comment */
				p ++;
				while(*p && *p != '\n') p++;
				break;
			}

			if(ch == ';') {
				/* sentence separator */
				p ++;
				break;
			}

			if(ch == '$') {
				/* variable reference */
				p ++;
				ref = p;
				p = skipuntilspace(p);
				if(ref == p) return TOK_PARSE_EMPTYVAR;
				tclose(&nextch, &p);

				vtokc = tok_getvar(t, ref, &vtokv);
				for(i=0; i<vtokc; i++)
					addtoken(t, vtokv[i]);
				continue;
			}

			if(ch == '"') {
				/* string literal */

				p ++;
				ref = p;
				q = ref;
				for(;;) {
					if(!*p) return TOK_PARSE_STRINGLIT;

					if(*p == '"') {
						*q = 0;
						addtoken(t, ref);
						p ++;
						break;
					}

					if(*p == '\\') {
						p ++;
						if(!*p) return TOK_PARSE_STRINGLIT;

						switch(*p) {
						case '\'':
						case '"':
						case '\\':
							*q++ = *p;
							break;
						case 'n':
							*q++ = '\n';
							break;
						case 't':
							*q++ = '\t';
							break;
						case 'b':
							*q++ = '\b';
							break;
						case 'x':
							{
								p ++;
								if(!*p) return TOK_PARSE_STRINGLIT;
								major = *p++;
								if(!*p) return TOK_PARSE_STRINGLIT;
								minor = *p;
								major = unhex(major);
								minor = unhex(minor);
								if(major == -1 || minor == -1) return TOK_PARSE_HEX;
								*q++ = major * 16 + minor;
							}
							break;
						default:
							return TOK_PARSE_ESCAPE;
						}

						p ++;
					} else {
						*q++ = *p++;
					}
				}

				continue;
			}

			ref = p;
			p = skipuntilspace(p);
			assert(ref != p);

			tclose(&nextch, &p);
			addtoken(t, ref);
		}

		/* dispatch sentence */
		if(t->tokc > 0) {
			int r = cb(t->tokc, t->tokv);
			if(r) return r;
		}
	}

	return 0;
}

/* set variable
 * all data is copied */
void tok_setvar(tok_t *t, const char *key, int tokc, char **tokv)
{
	ptrhash_entry_t *pe = ptrhash_getentry(t->vars, (void *) key);
	if(pe) {
		var_free(pe->value);
		if(tokc > 0) {
			pe->value = var_make(tokv, tokc);
		} else {
			void *oldkey = pe->key;
			ptrhash_del(t->vars, (void *) key);
			free(oldkey);
		}
	} else {
		if(tokc > 0) {
			ptrhash_set(t->vars, (void *) strdup(key), var_make(tokv, tokc));
		}
	}
}

/* get variable
 * returns token count */
int tok_getvar(tok_t *t, const char *key, char ***tokv)
{
	var_t *v;
	ptrhash_entry_t *pe = ptrhash_getentry(t->vars, (void *) key);
	if(!pe) return 0;

	v = pe->value;
	*tokv = v->tokv;
	return v->tokc;
}

/* return tokenizer error string */
const char *tok_error(int res)
{
	switch(res) {
	case 0:
		return "OK";
	case -1:
		return "Reference to empty variable name";
	case -2:
		return "String literal not closed";
	case -3:
		return "Escape sequence not recognized";
	case -4:
		return "Invalid hex literal in escape sequence";
	default:
		return NULL;
	}
}

/* return a concatenated string of tokens
 * uses a static buffer that's overwritten with consecutive calls */
const char *tok_print(int tc, char *tv[])
{
	static char *buf = NULL;
	static int alloc = 0;

	int i;
	int at = 0;

	for(i=0; i<tc; i++) {
		int l = strlen(tv[i]);
		if(at + l + 2 > alloc) {
			alloc = at + l + 2;
			buf = realloc(buf, alloc+16);
			if(!buf) abort();
		}

		buf[at] = ' ';
		strcpy(buf + at + 1, tv[i]);
		at += (l + 1);
	}

	if(!tc) return ""; else return buf;
}
