#include "util.h"

#include <stdlib.h>
#include <assert.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>

/* unordered list of pointers */

void ptrset_init(ptrset_t *set, ptrset_renum_call_t renum)
{
	set->n = 0;
	set->alloc = 0;
	set->p = NULL;
	set->renum = renum;
}

void ptrset_free(ptrset_t *set)
{
	free(set->p);
	set->n = 0;
	set->alloc = 0;
	set->p = NULL;
}

void ptrset_add(ptrset_t *set, void *ptr)
{
	if(set->n >= set->alloc) {
		if(!set->alloc) set->alloc = 8; else set->alloc *= 2;
		set->p = realloc(set->p, set->alloc * sizeof(set->p[0]));
		if(!set->p) abort();
	}
	int at = set->n;
	set->p[set->n++] = ptr;

	if(set->renum) set->renum(ptr, at);
}

void ptrset_del_idx(ptrset_t *set, int at)
{
	assert(at >= 0 && at < set->n);

	if(at == set->n - 1) {
		set->n --;
		return;
	}

	set->p[at] = set->p[set->n - 1];
	set->n --;

	if(set->renum) set->renum(set->p[at], at);
}

/* ordered list of pointers */

void ptrlist_init(ptrlist_t *l)
{
	l->first = NULL;
	l->last = NULL;
}

void ptrlist_free(ptrlist_t *l)
{
	ptrlist_el_t *p = l->first;
	while(p) {
		ptrlist_el_t *f = p;
		p = p->next;
		free(f);
	}
	l->first = NULL;
	l->last = NULL;
}

ptrlist_el_t * ptrlist_add_first(ptrlist_t *l, void *ptr)
{
	ptrlist_el_t *el = malloc(sizeof *el);
	if(!el) abort();
	el->ptr = ptr;

	el->prev = NULL;
	el->next = l->first;

	if(l->first) l->first->prev = el;

	l->first = el;
	if(!l->last) l->last = el;

	return el;
}

ptrlist_el_t * ptrlist_add_last(ptrlist_t *l, void *ptr)
{
	ptrlist_el_t *el = malloc(sizeof *el);
	if(!el) abort();
	el->ptr = ptr;

	el->prev = l->last;
	el->next = NULL;

	if(l->last) l->last->next = el;

	if(!l->first) l->first = el;
	l->last = el;

	return el;
}

void ptrlist_del_el(ptrlist_t *l, ptrlist_el_t *el)
{
	if(el->prev) {
		el->prev->next = el->next;
	}

	if(el->next) {
		el->next->prev = el->prev;
	}

	if(l->first == el) l->first = el->next;
	if(l->last == el) l->last = el->prev;

	free(el);
}

char *bufprintfbuf = NULL;
int bufprintfalloc = 0;

/* kinda-like-sprintf that allocates big enough buffer
 * writes into static buffer (bufprintfbuf),
 * that's overwritten with consecutive calls
 * returns length of message (excluding \0) */
int vbufprintf(const char *fmt, va_list ap)
{
	if(!bufprintfalloc) {
		bufprintfalloc = 128;
		bufprintfbuf = realloc(bufprintfbuf, bufprintfalloc);
		if(!bufprintfbuf) abort();
	}

	while(1) {
		va_list aq;
		va_copy(aq, ap);
		int n = vsnprintf(bufprintfbuf, bufprintfalloc, fmt, aq);
		va_end(aq);

		if(n > -1 && n < bufprintfalloc)
			return n;

		if(n > -1)
			bufprintfalloc = n + 1;
		else
			bufprintfalloc *= 2;

		bufprintfbuf = realloc(bufprintfbuf, bufprintfalloc + 512);
		if(!bufprintfbuf) abort();
	}
}

int bufprintf(const char *fmt, ...)
{
	int len;
	va_list ap;
	va_start(ap, fmt);
	len = vbufprintf(fmt, ap);
	va_end(ap);
	return len;
}

/* string assembling utility
 * works in a static buffer (separate from bufprintf's buffer) */
int strasm_len = 0;
char *strasm_buf = NULL;
static int strasm_alloc = 0;

void strasm_reset()
{
	strasm_len = 0;
}

void strasm_add(const char *str)
{
	int l = strlen(str);
	if(strasm_len + l + 1 > strasm_alloc) {
		strasm_alloc = strasm_len + l + 1;
		strasm_buf = realloc(strasm_buf, strasm_alloc);
		if(!strasm_buf) abort();
	}

	memcpy(strasm_buf + strasm_len, str, l + 1);
	strasm_len += l;
}
